import pytest

from tala.ddd.maker.ddd_maker import UnexpectedCharactersException

from .console_script_mixin import ConsoleScriptTestMixin


class TestCreateDDD(ConsoleScriptTestMixin):
    def test_create(self):
        self._when_creating_a_ddd(name="legal_name")
        self._then_result_is_successful()

    def _when_creating_a_ddd(self, name=None):
        self._create_ddd(name)

    def test_create_with_illegal_characters(self):
        self._when_creating_a_ddd_then_an_exception_is_raised(
            name="illegal-name",
            expected_exception=UnexpectedCharactersException,
            expected_pattern="Expected only alphanumeric ASCII and underscore characters in DDD name 'illegal-name', "
            "but found others"
        )

    def _when_creating_a_ddd_then_an_exception_is_raised(self, name, expected_exception, expected_pattern):
        with pytest.raises(expected_exception, match=expected_pattern):
            self._when_creating_a_ddd(name)



import json
import asyncio
from pathlib import Path

import prompt_toolkit
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.validation import Validator

from tala.cli.tdm.passivity_timer import PassivityTimer, PASSIVE
from tala.utils.tdm_client import TDMClient
from tala.utils.observable import Observer


class TDMCLI(object):
    def __init__(self, url, resume_session):
        self._client = TDMClient(url)
        self._tdm_session = self._create_session(resume_session)
        self._passivity_timer = PassivityTimer()
        self._passivity_timer.add_observer(PassivityTimerObserver(self))
        history = self._create_file_history()
        activity_detector = self._create_activity_detector_as_empty_validator()
        self._session = PromptSession(history=history, mouse_support=True, validator=activity_detector)

    def _create_session(self, resume_session):
        if resume_session:
            session = self.retrieve_session(resume_session)
        else:
            session = {}
        session["device_id"] = "tala"
        return session

    def retrieve_session(self, resume_session):
        if resume_session == "tala-session.state":
            return self.load_session_from_file()
        return json.loads(resume_session)

    def load_session_from_file(self):
        with open("tala-session.state") as file:
            return json.load(file)

    def _create_activity_detector_as_empty_validator(self):
        def on_active(text):
            if text:
                self._passivity_timer.stop()
            return True

        return Validator.from_callable(on_active)

    def stop(self):
        self._passivity_timer.stop()

    def run(self):
        for system_turn in self.system_turns():
            print(system_turn)

    def system_turns(self):
        if not self._tdm_session.get("session_id", None):
            greeting = self._request(self._client.start_session, self._tdm_session)
            yield greeting
        while True:
            message = self._session.prompt("U> ").rstrip()
            self._passivity_timer.stop()
            yield self._request(self._client.request_text_input, message, self._tdm_session)

    def _request(self, callable, *args, **kwargs):
        response = callable(*args, **kwargs)
        self._tdm_session = response["session"]
        expected_passivity = response["output"]["expected_passivity"]
        if expected_passivity is not None:
            self._passivity_timer.start(expected_passivity)
        system_utterance = response["output"]["utterance"]
        return f"S> {system_utterance}"

    def request_passivity(self):
        system_utterance = self._request(self._client.request_passivity, self._tdm_session)

        async def print_passivity_turn():
            def passivity_turn():
                print("U>")
                print(system_utterance)

            prompt_toolkit.application.run_in_terminal(passivity_turn)

        asyncio.run(print_passivity_turn())

    def _create_file_history(self):
        history_file = Path(".tala") / "history"
        if not history_file.parent.exists():
            history_file.parent.mkdir()
        if not history_file.exists():
            history_file.touch()
        return FileHistory(str(history_file))


class PassivityTimerObserver(Observer):
    def __init__(self, tdm_cli):
        self._tdm_cli = tdm_cli

    def update(self, value):
        if value == PASSIVE:
            self._tdm_cli.request_passivity()



from tala.model.error import OntologyError
from tala.model.predicate import Predicate
from tala.model.sort import CustomSort, IntegerSort, RealSort
from tala.testing.lib_test_case import LibTestCase


class PredicateTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def test_get_name(self):
        self.assertEqual("dest_city", self.predicate_dest_city.get_name())

    def test_create_invalid_predicate(self):
        with self.assertRaises(OntologyError):
            self.ontology.get_predicate("kalle")

    def test_getSort(self):
        predicate = self.ontology.get_predicate("dest_city")
        self.assertEqual(self.sort_city, predicate.sort)

    def test_feature_get_sort(self):
        predicate = self.ontology.get_predicate("dest_city_type")
        self.assertEqual(self.sort_city_type, predicate.sort)

    def test_is_feature_of_true(self):
        self.assertTrue(self.predicate_dest_city_type.is_feature_of(self.predicate_dest_city))

    def test_is_feature_of_false(self):
        self.assertFalse(self.predicate_dest_city.is_feature_of(self.predicate_dest_city_type))

    def test_allows_multiple_instances_false_by_default(self):
        arbirary_predicate = self.ontology.get_predicate("dest_city_type")
        self.assertFalse(arbirary_predicate.allows_multiple_instances())

    def test_allows_multiple_instances_true(self):
        multi_instance_predicate = self.ontology.get_predicate("passenger_type_to_add")
        self.assertTrue(multi_instance_predicate.allows_multiple_instances())

    def test_equality(self):
        predicate1 = Predicate(self.ontology_name, "dest_city", self._city_sort)
        predicate2 = Predicate(self.ontology_name, "dest_city", self._city_sort)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(predicate1, predicate2)

    def test_inequality_due_to_name(self):
        predicate1 = Predicate(self.ontology_name, "dest_city", self._city_sort)
        predicate2 = Predicate(self.ontology_name, "dept_city", self._city_sort)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(predicate1, predicate2)

    def test_inequality_due_to_sort(self):
        predicate1 = Predicate(self.ontology_name, "price", IntegerSort())
        predicate2 = Predicate(self.ontology_name, "price", RealSort())
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(predicate1, predicate2)

    def test_inequality_due_to_featurehood(self):
        predicate1 = Predicate(self.ontology_name, "dest_city_type", self._city_sort)
        predicate2 = Predicate(self.ontology_name, "dest_city_type", self._city_sort, feature_of_name="dest_city")
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(predicate1, predicate2)

    def test_inequality_due_to_multiple_instances(self):
        predicate1 = Predicate(self.ontology_name, "dest_city", self._city_sort, multiple_instances=False)
        predicate2 = Predicate(self.ontology_name, "dest_city", self._city_sort, multiple_instances=True)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(predicate1, predicate2)

    def test_inequality_due_to_non_predicate(self):
        non_predicate = "means_of_transport"
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.predicate_dest_city, non_predicate)

    def test_inequality_due_to_ontology_name(self):
        sort_of_an_ontology = CustomSort("an_ontology", "a_sort")
        predicate1 = Predicate("an_ontology", "dest_city", sort_of_an_ontology)
        sort_of_other_ontology = CustomSort("other_ontology", "a_sort")
        predicate2 = Predicate("other_ontology", "dest_city", sort_of_other_ontology)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(predicate1, predicate2)

    def test_unicode(self):
        self.assertEqual("dest_city", str(self.predicate_dest_city))

    def test_hashable(self):
        s = set()
        s.add(self.predicate_dest_city)



import pytest

from unittest.mock import call, Mock, patch, ANY

from tala.model.common import Modality
from tala.model.event_notification import EventNotification
from tala.model.input_hypothesis import InputHypothesis
from tala.model.interpretation import Interpretation, InterpretationWithoutUtterance
from tala.model.user_move import UserMove
from tala.testing.interactions import testcase
from tala.testing.interactions.named_test import InteractionTest
from tala.testing.interactions.testcase import InteractionTestingTestCase
from tala.testing.interactions.turn import RecognitionHypothesesTurn, UserInterpretationTurn, \
    SystemUtteranceTurn, UserPassivityTurn, NotifyStartedTurn, SystemMovesTurn
from tala.utils.tdm_client import TDMRuntimeException


class TestInteractionTestingTestCase:
    def setup_method(self):
        self._mock_tdm_client = None
        self._responses = []
        self._response_iterator = None

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_request_initial_system_output_if_test_has_initial_system_utterance(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_system_utterance("utterance")
        self.given_tdm_client_response(utterance="utterance")
        self.given_testcase()
        self.when_performing_test()
        self.then_starts_session()

    def given_mock_tdm_client(self, MockTDMClient):
        def side_effect(*args, **kwargs):
            if len(self._responses) > 1:
                if self._response_iterator is None:
                    self._response_iterator = iter(self._responses)
                return next(self._response_iterator)
            if len(self._responses) > 0:
                return self._responses[0]
            return self._create_tdm_client_response()

        self._mock_tdm_client = MockTDMClient.return_value
        self._mock_tdm_client.request_text_input.side_effect = side_effect
        self._mock_tdm_client.request_speech_input.side_effect = side_effect
        self._mock_tdm_client.request_semantic_input.side_effect = side_effect
        self._mock_tdm_client.request_passivity.side_effect = side_effect
        self._mock_tdm_client.request_event_notification.side_effect = side_effect
        self._mock_tdm_client.start_session.side_effect = side_effect

    def then_starts_session(self):
        self._mock_tdm_client.start_session.assert_called_with(session_data={"device_id": "tala"})

    def given_tdm_client_response(self, *args, **kwargs):
        self._responses.append(self._create_tdm_client_response(*args, **kwargs))  # yapf: disable

    def _create_tdm_client_response(
        self,
        utterance="mock_system_utterance",
        moves=None,
        expected_passivity=None,
        actions=None,
        nlu_result="mock_nlu_result",
        context="mock_context",
        selected_interpretation="mock_selected_interpretation",
        expected_input="mock_expected_input"
    ):
        return {
            "version": "3.4",
            "session": {
                "session_id": "mock-session-id"
            },
            "output": {
                "moves": moves or [],
                "utterance": utterance,
                "expected_passivity": expected_passivity,
                "actions": actions or []
            },
            "nlu_result": nlu_result,
            "context": context,
            "selected_interpretation": selected_interpretation,
            "expected_input": expected_input
        }

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_request_initial_system_output_if_test_has_initial_user_utterance(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_user_utterance("mockup_utterance")
        self.given_testcase()
        self.when_performing_test()
        self.then_starts_session()

    def then_request_is_not_enqueued(self, request):
        unexpected_call = call(request)
        actual_calls = self._mock_tdm_client.enqueue_request.call_args_list
        if unexpected_call in actual_calls:
            raise AssertionError(
                "Expected the request '%s' not to be enqueued but found it in the list %s" % (request, actual_calls)
            )

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_user_utterance_with_explicit_score(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_user_utterance("mockup_utterance", "1.0")
        self.given_testcase()
        self.when_performing_test()
        self.then_requests_speech_input([InputHypothesis("mockup_utterance", 1.0)])

    def then_requests_speech_input(self, expected_hypotheses):
        self._mock_tdm_client.request_speech_input.assert_called_with(expected_hypotheses, ANY)

    def given_mock_backends_return_frontend_simulator(self):
        self._mock_tdm_client = self.mock_backend_factory.create.return_value
        self._mock_tdm_client._get_module_instance.return_value = self.mock_frontend_simulator

    def given_test_with_user_utterance(self, utterance, score_or_confidence_level="1.0", line_number=1):
        turn = RecognitionHypothesesTurn([(utterance, score_or_confidence_level)], line_number)
        self.test = self._create_mock_test([turn])

    def given_testcase(self):
        mock_domain_orchestration = Mock()
        self.testcase = InteractionTestingTestCase(self.test, mock_domain_orchestration)

    def when_performing_test(self):
        self.testcase.setUp()
        self.testcase.perform()
        self.testcase.tearDown()

    def _create_mock_test(self, turns):
        test = Mock(spec=InteractionTest)
        test.turns = turns
        test.name = "mock_test_name"
        test.filename = "mock_file_name"
        return test

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_user_utterance_with_confidence_level(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_user_utterance("mockup_utterance", "$ACKNOWLEDGE")
        self.given_testcase()
        self.when_performing_test_then_exception_is_raised(
            NotImplementedError, match="Explicit names of confidence levels are not supported in this version."
        )

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_system_utterance_that_matches(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_system_utterance("mockup_utterance")
        self.given_tdm_client_response(utterance="mockup_utterance")
        self.given_testcase()
        self.when_performing_test()
        self.then_test_is_successful()

    def then_test_is_successful(self):
        pass

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_system_utterance_not_matching_output(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_system_utterance("utterance")
        self.given_tdm_client_response(utterance="another")
        self.given_testcase()
        self.when_performing_test_then_it_fails_with_message(
            """On line 1 of mock_file_name,

expected:
  S> utterance

but got:
  S> another"""
        )

    def when_performing_test_then_it_fails_with_message(self, expected_message):
        with pytest.raises(AssertionError, match=expected_message):
            self.when_performing_test()

    def given_test_with_system_utterance(self, utterance, line_number=1):
        turn = SystemUtteranceTurn(utterance, line_number)
        self.test = self._create_mock_test([turn])

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_system_moves_that_matches(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_system_moves(["mock_move"])
        self.given_tdm_client_response(moves=["mock_move"])
        self.given_testcase()
        self.when_performing_test()
        self.then_test_is_successful()

    def given_test_with_system_moves(self, moves, line_number=1):
        turn = SystemMovesTurn(moves, line_number)
        self.test = self._create_mock_test([turn])

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_system_moves_not_matching_output(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_system_moves(["a_move"])
        self.given_tdm_client_response(moves=["another_move"])
        self.given_testcase()
        self.when_performing_test_then_it_fails_with_message(
            r"""On line 1 of mock_file_name,

expected:
  S> \["a_move"\]

but got:
  S> \["another_move"\]"""
        )

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_passivity_after_system_utterance(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_system_utterances_and_passivity(
            utterance_before_passivity="utterance before passivity",
            utterance_after_passivity="utterance after passivity"
        )
        self.given_tdm_client_response(utterance="utterance before passivity", expected_passivity=5.0)
        self.given_tdm_client_response(utterance="utterance after passivity")
        self.given_testcase()
        self.when_performing_test()
        self.then_requests_passivity()

    def then_requests_passivity(self):
        self._mock_tdm_client.request_passivity.assert_called_with(ANY)

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_user_moves(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_user_interpretation(["semantic_expression_1", "semantic_expression_2"])
        self.given_tdm_client_response(utterance="an utterance")
        self.given_testcase()
        self.when_performing_test()
        self.then_requests_semantic_input([
            Interpretation([UserMove("semantic_expression_1", 1.0, 1.0),
                            UserMove("semantic_expression_2", 1.0, 1.0)], Modality.OTHER)
        ])

    def then_requests_semantic_input(self, expected_interpretations):
        self._mock_tdm_client.request_semantic_input.assert_called_with(expected_interpretations, ANY)

    def given_test_with_user_interpretation(self, moves, line_number=1, utterance=None, modality=None):
        turn = UserInterpretationTurn(moves, line_number, modality, utterance)
        self.test = self._create_mock_test([turn])

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_json_user_moves(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_user_interpretation([{
            "semantic_expression": "semantic_expression_1",
            "understanding_confidence": 0.12,
            "perception_confidence": 0.11,
        }, {
            "semantic_expression": "semantic_expression_2",
            "perception_confidence": 0.21,
            "understanding_confidence": 0.22
        }])
        self.given_tdm_client_response(utterance="an utterance")
        self.given_testcase()
        self.when_performing_test()
        self.then_requests_semantic_input([
            InterpretationWithoutUtterance(
                moves=[UserMove("semantic_expression_1", 0.11, 0.12),
                       UserMove("semantic_expression_2", 0.21, 0.22)],
                modality=Modality.OTHER,
            )
        ])

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_json_user_interpretation(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_user_interpretation(
            moves=[{
                "semantic_expression": "semantic_expression_1",
                "understanding_confidence": 0.12,
                "perception_confidence": 0.11,
            }, {
                "semantic_expression": "semantic_expression_2",
                "perception_confidence": 0.21,
                "understanding_confidence": 0.22
            }],
            utterance="mock utterance",
            modality="speech"
        )
        self.given_tdm_client_response(utterance="an utterance")
        self.given_testcase()
        self.when_performing_test()
        self.then_requests_semantic_input([
            Interpretation(
                moves=[UserMove("semantic_expression_1", 0.11, 0.12),
                       UserMove("semantic_expression_2", 0.21, 0.22)],
                modality=Modality.SPEECH,
                utterance="mock utterance",
            )
        ])

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_notify_started(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_notify_started("mock_action", {})
        self.given_testcase()
        self.when_performing_test()
        self.then_requests_event_notification(EventNotification("mock_action", EventNotification.STARTED, {}))

    def then_requests_event_notification(self, expected_event_notifications):
        self._mock_tdm_client.request_event_notification.assert_called_with(expected_event_notifications, ANY)

    def given_test_with_notify_started(self, action, parameters, line_number=1):
        turn = NotifyStartedTurn(action, parameters, line_number)
        self.test = self._create_mock_test([turn])

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_failure_when_no_response_to_user_passivity(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_system_utterances_and_passivity(
            utterance_before_passivity="utterance before passivity",
            line_number_of_passivity=2,
            utterance_after_passivity="utterance after passivity",
            line_number_of_utterance_after_passivity=3
        )
        self.given_tdm_client_response(utterance="utterance before passivity", expected_passivity=None)
        self.given_testcase()
        self.when_performing_test_then_it_fails_with_message(
            """

On line 3 of mock_file_name,
expected:
  S> utterance after passivity

in response to
  U>
on line 2

but the system didn't expect user passivity."""
        )

    def when_performing_test_then_exception_is_raised(self, *args, **kwargs):
        with pytest.raises(*args, **kwargs):
            self.when_performing_test()

    def given_test_with_system_utterances_and_passivity(
        self,
        utterance_before_passivity="utterance before passivity",
        line_number_of_utterance_before_passivity=1,
        line_number_of_passivity=2,
        utterance_after_passivity="utterance after passivity",
        line_number_of_utterance_after_passivity=3
    ):
        turns = [
            SystemUtteranceTurn(utterance_before_passivity, line_number_of_utterance_before_passivity),
            UserPassivityTurn(line_number_of_passivity),
            SystemUtteranceTurn(utterance_after_passivity, line_number_of_utterance_after_passivity)
        ]
        self.test = self._create_mock_test(turns)

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_system_returns_nothing_when_expecting_system_utterance(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_system_utterance("utterance")
        self.given_error("Error message.")
        self.given_testcase()
        self.when_performing_test_then_exception_is_raised(TDMRuntimeException, match="Error message.")

    def given_error(self, message):
        side_effect = TDMRuntimeException(message)
        self._mock_tdm_client.request_text_input.side_effect = side_effect
        self._mock_tdm_client.request_speech_input.side_effect = side_effect
        self._mock_tdm_client.request_semantic_input.side_effect = side_effect
        self._mock_tdm_client.request_passivity.side_effect = side_effect
        self._mock_tdm_client.request_event_notification.side_effect = side_effect
        self._mock_tdm_client.start_session.side_effect = side_effect

    @patch(f"{testcase.__name__}.TDMClient", autospec=True)
    def test_system_errors_when_expecting_system_utterance(self, MockTDMClient):
        self.given_mock_tdm_client(MockTDMClient)
        self.given_test_with_system_utterance("utterance")
        self.given_error("an error")
        self.given_testcase()
        self.when_performing_test_then_exception_is_raised(TDMRuntimeException, match="an error")



from tala.testing.interaction_test_noveau.interaction_tester import InteractionTester


class InteractionTestExecutorBase:
    def run_ng_interaction_test(self, testcase, port=None):
        self._given_testcase(testcase)
        self._when_running_testcase(port)

    def _given_testcase(self, testcase):
        self._testcase = testcase

    def _when_running_testcase(self, port):
        tester = InteractionTester(port)
        self._result = tester.run_testcase(self._testcase)


class InteractionTestExecutorNoAssert(InteractionTestExecutorBase):
    @property
    def result(self):
        return self._result


class InteractionTestExecutorAssert(InteractionTestExecutorBase):
    def _when_running_testcase(self, port):
        super()._when_running_testcase(port)
        self._then_test_is_succesful()

    def _then_test_is_succesful(self):
        print(self._result["transcript"])
        assert self._result["success"], self._result["failure_description"]



from mock import Mock

from tala.ddd.services.service_interface import ServiceInterface, ServiceActionInterface, ServiceValidatorInterface, DeviceModuleTarget


class DddCompilerTestCase(object):
    DDD_NAME = "mockup_ddd_name"

    def setup_method(self):
        self._service_interface = None
        self.mock_service_target = self._create_mock_service_target()

    def _create_mock_service_target(self):
        mock_target = Mock(spec=DeviceModuleTarget)
        mock_target.is_frontend = False
        return mock_target

    def _given_service_interface_with_validity(self, validity_name):
        self._create_service_interface(
            validators=[ServiceValidatorInterface(validity_name, self.mock_service_target, parameters=[])]
        )

    def _create_service_interface(self, actions=None, queries=None, validators=None, entity_recognizers=None):
        def has_action(name):
            return any(name == action.name for action in self._service_interface.actions)

        self._service_interface = Mock(spec=ServiceInterface)
        self._service_interface.has_action.side_effect = has_action
        self._service_interface.actions = actions or []
        self._service_interface.queries = queries or []
        self._service_interface.validators = validators or []
        self._service_interface.entity_recognizers = entity_recognizers or []

    def _given_service_interface_without_actions(self):
        self._create_service_interface()

    def _given_service_interface_with_action(self, action_name):
        self._create_service_interface(
            actions=[ServiceActionInterface(action_name, self.mock_service_target, parameters=[], failure_reasons=[])]
        )



from mock import patch, ANY

from tala.cli.tdm import passivity_timer
from tala.cli.tdm.passivity_timer import PassivityTimer
from tala.utils.observable import Observer


class TestPassivityTimer(object):
    def setup_method(self):
        self._passivity_timer = PassivityTimer()
        self._MockTimer = None
        self._mocked_timer = None
        self._observer = None

    @patch("{}.Timer".format(passivity_timer.__name__, autospec=True))  # noqa: F522
    def test_start(self, MockTimer):
        self._given_mocked_timer(MockTimer)
        self._when_starting_timer_with(5.0)
        self._then_timer_was_started_with(5.0)

    def _given_mocked_timer(self, MockTimer):
        self._MockTimer = MockTimer
        self._mocked_timer = MockTimer.return_value

    def _when_starting_timer_with(self, duration):
        self._start_timer(duration)

    def _start_timer(self, duration=1.0):
        self._passivity_timer.start(duration)

    def _then_timer_was_started_with(self, expected_duration):
        self._MockTimer.assert_called_once_with(expected_duration, function=ANY)
        self._mocked_timer.start.assert_called_once()

    @patch("{}.Timer".format(passivity_timer.__name__, autospec=True))  # noqa: F522
    def test_stop(self, MockTimer):
        self._given_mocked_timer(MockTimer)
        self._given_started_timer()
        self._when_stopping_timer()
        self._then_timer_was_stopped()

    def _given_started_timer(self):
        self._start_timer()

    def _when_stopping_timer(self):
        self._passivity_timer.stop()

    def _then_timer_was_stopped(self):
        self._mocked_timer.cancel.assert_called_once()

    @patch("{}.Timer".format(passivity_timer.__name__, autospec=True))  # noqa: F522
    def test_stop_timer_when_not_started(self, MockTimer):
        self._given_mocked_timer(MockTimer)
        self._when_stopping_timer()
        self._then_no_exceptions_occur()

    def _then_no_exceptions_occur(self):
        pass

    @patch("{}.Timer".format(passivity_timer.__name__, autospec=True))  # noqa: F522
    def test_timer_updates_observer(self, MockTimer):
        self._given_mocked_timer(MockTimer)
        self._given_observer_added()
        self._given_started_timer()
        self._when_timer_triggers()
        self._then_observer_is_updated()

    def _given_observer_added(self):
        class PassivityTimerObserver(Observer):
            def __init__(self):
                self.updated_with = None

            def update(self, value):
                self.updated_with = value

        self._observer = PassivityTimerObserver()
        self._passivity_timer.add_observer(self._observer)

    def _when_timer_triggers(self):
        args, kwargs = self._MockTimer.call_args
        trigger_callback = kwargs["function"]
        trigger_callback()

    def _then_observer_is_updated(self):
        assert self._observer.updated_with == passivity_timer.PASSIVE



class Turn(object):
    def __init__(self, line_number):
        self._line_number = line_number

    @property
    def is_user_input_turn(self):
        return False

    @property
    def is_system_output_turn(self):
        return False

    @property
    def is_user_passivity_turn(self):
        return False

    @property
    def is_event_turn(self):
        return False

    @property
    def line_number(self):
        return self._line_number


class UserInputTurn(Turn):
    def __init__(self, line_number):
        super(UserInputTurn, self).__init__(line_number)

    @property
    def is_user_input_turn(self):
        return True

    @property
    def is_recognition_hypotheses_turn(self):
        return False

    @property
    def is_user_interpretation_turn(self):
        return False

    @property
    def is_user_semantic_input_turn(self):
        return False

    @property
    def is_user_passivity_turn(self):
        return False


class RecognitionHypothesesTurn(UserInputTurn):
    def __init__(self, hypotheses, line_number):
        super(RecognitionHypothesesTurn, self).__init__(line_number)
        self._hypotheses = hypotheses

    @property
    def is_recognition_hypotheses_turn(self):
        return True

    @property
    def hypotheses(self):
        return self._hypotheses


class UserInterpretationTurn(UserInputTurn):
    def __init__(self, moves, line_number, modality=None, utterance=None):
        super(UserInterpretationTurn, self).__init__(line_number)
        self._moves = moves
        self._modality = modality
        self._utterance = utterance

    @property
    def is_user_interpretation_turn(self):
        return True

    @property
    def moves(self):
        return self._moves

    @property
    def modality(self):
        return self._modality

    @property
    def utterance(self):
        return self._utterance


class UserSemanticInputTurn(UserInputTurn):
    def __init__(self, semantic_input, line_number):
        super().__init__(line_number)
        self._semantic_input = semantic_input

    @property
    def is_user_semantic_input_turn(self):
        return True

    @property
    def semantic_input(self):
        return self._semantic_input


class UserPassivityTurn(Turn):
    @property
    def is_user_passivity_turn(self):
        return True


class SystemOutputTurn(Turn):
    @property
    def output(self):
        raise NotImplementedError()

    @property
    def is_system_output_turn(self):
        return True

    @property
    def is_system_utterance_turn(self):
        return False

    @property
    def is_system_moves_turn(self):
        return False


class SystemUtteranceTurn(SystemOutputTurn):
    def __init__(self, utterance, line_number):
        super(SystemOutputTurn, self).__init__(line_number)
        self._utterance = utterance

    @property
    def output(self):
        return self.utterance

    @property
    def is_system_utterance_turn(self):
        return True

    @property
    def utterance(self):
        return self._utterance

    def __str__(self):
        return "S>"


class SystemMovesTurn(SystemOutputTurn):
    def __init__(self, moves, line_number):
        super(SystemOutputTurn, self).__init__(line_number)
        self._moves = moves

    @property
    def output(self):
        return self.moves

    @property
    def is_system_moves_turn(self):
        return True

    @property
    def moves(self):
        return self._moves


class EventTurn(Turn):
    @property
    def is_event_turn(self):
        return True

    @property
    def is_notify_started_turn(self):
        return False


class NotifyStartedTurn(EventTurn):
    def __init__(self, action, parameters, line_number):
        super(NotifyStartedTurn, self).__init__(line_number)
        self._action = action
        self._parameters = parameters

    @property
    def is_notify_started_turn(self):
        return True

    @property
    def action(self):
        return self._action

    @property
    def parameters(self):
        return self._parameters



import json

from mock import patch, ANY

from tala.cli.tdm import tdm_cli
from tala.cli.tdm.passivity_timer import PASSIVE
from tala.cli.tdm.tdm_cli import TDMCLI
from tala.utils.tdm_client import PROTOCOL_VERSION


class TestTDMCLI(object):
    def setup_method(self):
        self._mocked_tdm_client = None
        self._MockPromptSession = None
        self._mocked_prompt_session = None
        self._tdm_cli = None
        self._system_turns = None
        self._system_turn = None
        self._MockFileHistory = None
        self._mocked_file_history = None
        self._mocked_passivity_timer = None
        self._mocked_application = None
        self._capsys = None

    @patch("{}.PromptSession".format(tdm_cli.__name__), autospec=True)
    @patch("{}.TDMClient".format(tdm_cli.__name__), autospec=True)
    def test_first_system_turn_starts_session(self, MockTDMClient, MockPromptSession):
        self._given_mocked_tdm_client(MockTDMClient)
        self._given_mocked_prompt_session(MockPromptSession)
        self._given_created_tdm_cli()
        self._given_initiated_system_turns()
        self._when_getting_system_turn()
        self._then_session_is_started()

    def _given_mocked_tdm_client(self, MockTDMClient):
        self._mocked_tdm_client = MockTDMClient.return_value
        self._mocked_tdm_client.start_session.return_value = self._create_tdm_client_response()
        self._mocked_tdm_client.request_text_input.return_value = self._create_tdm_client_response()
        self._mocked_tdm_client.request_passivity.return_value = self._create_tdm_client_response()

    def _create_tdm_client_response(self, system_utterance="mock-system-utterance", expected_passivity="null"):
        return json.loads(
            '''{{
  "version": {version},
  "session": {{
    "session_id": "mock-session-id"
  }},
  "output": {{
    "utterance": "{system_utterance}",
    "expected_passivity": {expected_passivity},
    "actions": []
  }},
  "nlu_result": {{
    "selected_utterance": "call John",
    "confidence": 0.88
  }},
  "context": {{
    "active_ddd": "my_ddd",
    "facts": {{}},
    "language": "eng"
  }}
}}'''.format(version=PROTOCOL_VERSION, system_utterance=system_utterance, expected_passivity=expected_passivity)
        )

    def _given_mocked_prompt_session(self, MockPromptSession):
        self._MockPromptSession = MockPromptSession
        self._mocked_prompt_session = MockPromptSession.return_value

    def _given_created_tdm_cli(self):
        self._tdm_cli = self._create_tdm_cli()

    def _create_tdm_cli(self, url="mocked-url"):
        return TDMCLI(url, None)

    def _given_initiated_system_turns(self):
        self._system_turns = self._tdm_cli.system_turns()

    def _when_getting_system_turn(self):
        self._system_turn = next(self._system_turns)

    def _then_session_is_started(self):
        self._mocked_tdm_client.start_session.assert_called_once()

    @patch("{}.PromptSession".format(tdm_cli.__name__), autospec=True)
    @patch("{}.TDMClient".format(tdm_cli.__name__), autospec=True)
    def test_prompt_started_after_session_is_started(self, MockTDMClient, MockPromptSession):
        self._given_mocked_tdm_client(MockTDMClient)
        self._given_mocked_prompt_session(MockPromptSession)
        self._given_created_tdm_cli()
        self._given_initiated_system_turns()
        self._given_session_is_started()
        self._when_getting_system_turn()
        self._then_prompt_is_started_with("U> ")

    def _given_session_is_started(self):
        next(self._system_turns)
        self._mocked_tdm_client.start_session.assert_called_once()

    def _then_prompt_is_started_with(self, expected_message):
        self._mocked_prompt_session.prompt.assert_called_once_with(expected_message)

    @patch("{}.PromptSession".format(tdm_cli.__name__), autospec=True)
    @patch("{}.TDMClient".format(tdm_cli.__name__), autospec=True)
    @patch("{}.FileHistory".format(tdm_cli.__name__), autospec=True)
    def test_file_history_is_used(self, MockFileHistory, MockTDMClient, MockPromptSession):
        self._given_mocked_file_history(MockFileHistory)
        self._given_mocked_tdm_client(MockTDMClient)
        self._given_mocked_prompt_session(MockPromptSession)
        self._when_creating_tdm_cli()
        self._then_file_history_is_used()

    def _given_mocked_file_history(self, MockFileHistory):
        self._MockFileHistory = MockFileHistory
        self._mocked_file_history = self._MockFileHistory.return_value

    def _when_creating_tdm_cli(self):
        self._create_tdm_cli()

    def _then_file_history_is_used(self):
        self._MockPromptSession.assert_called_once_with(
            history=self._mocked_file_history, mouse_support=True, validator=ANY
        )

    @patch("{}.PromptSession".format(tdm_cli.__name__), autospec=True)
    @patch("{}.TDMClient".format(tdm_cli.__name__), autospec=True)
    @patch("{}.FileHistory".format(tdm_cli.__name__), autospec=True)
    def test_file_history_is_stored_in_local_tala_folder(self, MockFileHistory, MockTDMClient, MockPromptSession):
        self._given_mocked_file_history(MockFileHistory)
        self._given_mocked_tdm_client(MockTDMClient)
        self._given_mocked_prompt_session(MockPromptSession)
        self._when_creating_tdm_cli()
        self._then_file_history_is_stored_in(".tala/history")

    def _then_file_history_is_stored_in(self, expected_path):
        self._MockFileHistory.assert_called_once_with(expected_path)

    @patch("{}.PromptSession".format(tdm_cli.__name__), autospec=True)
    @patch("{}.TDMClient".format(tdm_cli.__name__), autospec=True)
    def test_system_utterance_from_tdm_client_is_used_as_system_turn(self, MockTDMClient, MockPromptSession):
        self._given_mocked_tdm_client(MockTDMClient)
        self._given_mocked_tdm_client_returns_utterance_on_start_session("system-utterance")
        self._given_mocked_prompt_session(MockPromptSession)
        self._given_created_tdm_cli()
        self._given_initiated_system_turns()
        self._when_getting_system_turn()
        self._then_system_turn_is("S> system-utterance")

    def _given_mocked_tdm_client_returns_utterance_on_start_session(self, utterance):
        self._mocked_tdm_client.start_session.return_value = \
            self._create_tdm_client_response(system_utterance=utterance)

    def _then_system_turn_is(self, turn):
        assert self._system_turn == turn

    @patch("{}.PromptSession".format(tdm_cli.__name__), autospec=True)
    @patch("{}.TDMClient".format(tdm_cli.__name__), autospec=True)
    @patch("{}.PassivityTimer".format(tdm_cli.__name__), autospec=True)
    def test_passivity_timer_is_started(self, MockPassivityTimer, MockTDMClient, MockPromptSession):
        self._given_mocked_passivity_timer(MockPassivityTimer)
        self._given_mocked_tdm_client(MockTDMClient)
        self._given_mocked_tdm_client_expects_passivity_when_session_is_started(expected_passivity=5.0)
        self._given_mocked_prompt_session(MockPromptSession)
        self._given_created_tdm_cli()
        self._given_initiated_system_turns()
        self._when_getting_system_turn()
        self._then_passivity_timer_is_started_with(5.0)

    def _given_mocked_passivity_timer(self, MockPassivityTimer):
        self._mocked_passivity_timer = MockPassivityTimer.return_value

    def _given_mocked_tdm_client_expects_passivity_when_session_is_started(self, expected_passivity):
        self._mocked_tdm_client.start_session.return_value = \
            self._create_tdm_client_response(expected_passivity=expected_passivity)

    def _then_passivity_timer_is_started_with(self, expected_duration):
        self._mocked_passivity_timer.start.assert_called_once_with(expected_duration)

    @patch("{}.PromptSession".format(tdm_cli.__name__), autospec=True)
    @patch("{}.TDMClient".format(tdm_cli.__name__), autospec=True)
    @patch("{}.PassivityTimer".format(tdm_cli.__name__), autospec=True)
    def test_passivity_timer_triggers_call_to_tdm_client(self, MockPassivityTimer, MockTDMClient, MockPromptSession):
        self._given_mocked_passivity_timer(MockPassivityTimer)
        self._given_mocked_tdm_client(MockTDMClient)
        self._given_mocked_prompt_session(MockPromptSession)
        self._given_created_tdm_cli()
        self._given_initiated_system_turns()
        self._when_passivity_timer_triggers()
        self._then_passivity_is_called_in_tdm_client()

    def _when_passivity_timer_triggers(self):
        args, kwargs = self._mocked_passivity_timer.add_observer.call_args
        observer, = args
        observer.update(PASSIVE)

    def _then_passivity_is_called_in_tdm_client(self):
        self._mocked_tdm_client.request_passivity.assert_called_once()

    @patch("{}.PromptSession".format(tdm_cli.__name__), autospec=True)
    @patch("{}.TDMClient".format(tdm_cli.__name__), autospec=True)
    @patch("{}.prompt_toolkit.application".format(tdm_cli.__name__), autospec=True)
    @patch("{}.PassivityTimer".format(tdm_cli.__name__), autospec=True)
    def test_system_utterance_from_passivity_call_to_tdm_client_is_used_as_system_utterance(
        self, MockPassivityTimer, mock_application, MockTDMClient, MockPromptSession, capsys
    ):
        self._given_mocked_passivity_timer(MockPassivityTimer)
        self._given_mocked_application(mock_application)
        self._given_mocked_tdm_client(MockTDMClient)
        self._given_mocked_tdm_client_returns_utterance_on_passivity("system-utterance")
        self._given_mocked_prompt_session(MockPromptSession)
        self._given_stdout_is_being_captured(capsys)
        self._given_created_tdm_cli()
        self._given_initiated_system_turns()
        self._given_session_is_started()
        self._when_passivity_timer_triggers()
        self._then_prompt_toolkit_is_instructed_to_print("U>\nS> system-utterance\n")

    def _given_mocked_application(self, mock_application):
        self._mocked_application = mock_application

    def _given_mocked_tdm_client_returns_utterance_on_passivity(self, utterance):
        self._mocked_tdm_client.request_passivity.return_value = \
            self._create_tdm_client_response(system_utterance=utterance)

    def _given_stdout_is_being_captured(self, capsys):
        self._capsys = capsys

    def _then_prompt_toolkit_is_instructed_to_print(self, expected_string):
        def get_print_callable_from_prompt_toolkit():
            self._mocked_application.run_in_terminal.assert_called_once()
            args, kwargs = self._mocked_application.run_in_terminal.call_args
            print_callable, = args
            return print_callable

        def actual_stdout():
            captured = self._capsys.readouterr()
            return captured.out

        printer = get_print_callable_from_prompt_toolkit()
        printer()
        assert actual_stdout() == expected_string

    @patch("{}.PromptSession".format(tdm_cli.__name__), autospec=True)
    @patch("{}.TDMClient".format(tdm_cli.__name__), autospec=True)
    @patch("{}.Validator".format(tdm_cli.__name__), autospec=True)
    @patch("{}.PassivityTimer".format(tdm_cli.__name__), autospec=True)
    def test_passivity_timer_stopped_when_user_becomes_active(
        self, MockPassivityTimer, MockValidator, MockTDMClient, MockPromptSession
    ):
        self._given_mocked_passivity_timer(MockPassivityTimer)
        self._given_mocked_user_activity_detector(MockValidator)
        self._given_mocked_tdm_client(MockTDMClient)
        self._given_mocked_prompt_session(MockPromptSession)
        self._given_created_tdm_cli()
        self._when_user_activity_is_detected()
        self._then_passivity_timer_is_stopped()

    def _given_mocked_user_activity_detector(self, MockValidator):
        self._mocked_activity_detector = MockValidator.from_callable

    def _when_user_activity_is_detected(self):
        def get_detector_callable():
            args, kwargs = self._mocked_activity_detector.call_args
            activity_detector_callable, = args
            return activity_detector_callable

        detector = get_detector_callable()
        detector("mock-input")

    def _then_passivity_timer_is_stopped(self):
        self._mocked_passivity_timer.stop.assert_called_once()

    @patch("{}.PromptSession".format(tdm_cli.__name__), autospec=True)
    @patch("{}.TDMClient".format(tdm_cli.__name__), autospec=True)
    def test_endurance(self, MockTDMClient, MockPromptSession):
        self._given_mocked_tdm_client(MockTDMClient)
        self._given_mocked_tdm_client_returns_utterance_on_request_text_input("system-utterance")
        self._given_mocked_prompt_session(MockPromptSession)
        self._given_created_tdm_cli()
        self._given_initiated_system_turns()
        self._given_session_is_started()
        self._when_user_and_system_are_taking_turns_for_a_long_while()
        self._then_system_turn_is("S> system-utterance")

    def _given_mocked_tdm_client_returns_utterance_on_request_text_input(self, utterance):
        self._mocked_tdm_client.request_text_input.return_value = self._create_tdm_client_response(
            system_utterance=utterance
        )

    def _when_user_and_system_are_taking_turns_for_a_long_while(self):
        for i in range(100):
            next(self._system_turns)
        self._system_turn = next(self._system_turns)



import copy
import re
import warnings

from tala.model.error import OntologyError
from tala.model.individual import Individual, NegativeIndividual
from tala.model.lambda_abstraction import LambdaAbstractedPredicateProposition
from tala.model.proposition import PredicateProposition
from tala.model.question import YesNoQuestion, WhQuestion
from tala.model.action import Action
from tala.model.sort import DomainSort, Sort
from tala.model.polarity import Polarity
from tala.utils.as_json import AsJSONMixin
from tala.utils.unique import unique


class IndividualExistsException(Exception):
    pass


class SortDoesNotExistException(Exception):
    pass


class ActionDoesNotExistException(Exception):
    pass


class PredicateDoesNotExistException(Exception):
    pass


class InvalidIndividualName(Exception):
    pass


class AmbiguousNamesException(Exception):
    pass


individual_name_matcher = re.compile(r"^[a-zA-Z0-9_\-\:]{1,}$")
SEMANTIC_OBJECT_TYPE = "ontology"


class DddOntology:
    pass


class Ontology(AsJSONMixin):
    DEFAULT_ACTIONS = {"top", "up", "how"}

    def __init__(self, name, sorts, predicates, individuals, actions):
        super(Ontology, self).__init__()
        self._name = name
        self._sorts = self._set_to_dict(sorts)
        self._predicates = self._set_to_dict(predicates)
        self._individuals = individuals
        self._original_individuals = copy.deepcopy(individuals)
        self._actions = self.DEFAULT_ACTIONS.union(actions)
        self._add_default_sorts()
        self._add_builtin_sorts_for_predicates()
        self._validate_individuals()
        self._validate_predicates()
        self._check_predicate_action_integrity()

    def as_dict(self):
        json = super(Ontology, self).as_dict()
        json["_original_individuals"] = "<skipped>"
        json["semantic_object_type"] = SEMANTIC_OBJECT_TYPE
        json["_sorts"] = [sort.as_dict() for sort in self._sorts.values() if not sort.is_builtin()]
        return json

    def reset(self):
        self._individuals = copy.deepcopy(self._original_individuals)

    @property
    def name(self):
        return self._name

    def _set_to_dict(self, set_):
        dict_ = {}
        for item in set_:
            dict_[item.get_name()] = item
        return dict_

    def _validate_individuals(self):
        for name, sort in list(self._individuals.items()):
            self._validate_individual_name(name)
            self._validate_individual_sort(name, sort)

    def _validate_individual_name(self, name):
        if not individual_name_matcher.match(name):
            raise InvalidIndividualName("invalid individual name %r" % name)

    def _validate_individual_sort(self, name, sort):
        if sort not in list(self._sorts.values()):
            raise OntologyError("individual '%s' has unknown sort '%s' (%s)" % (name, sort.get_name(), self.__dict__))

    def _add_default_sorts(self):
        self._sorts["domain"] = DomainSort()

    def _add_builtin_sorts_for_predicates(self):
        for predicate in list(self._predicates.values()):
            self._add_builtin_sorts_for_predicate(predicate)

    def _add_builtin_sorts_for_predicate(self, predicate):
        sort = predicate.sort
        if sort.is_builtin() and sort not in self._sorts:
            self._sorts[sort.get_name()] = sort

    def _validate_predicates(self):
        for predicate in list(self._predicates.values()):
            if predicate.sort not in list(self._sorts.values()):
                raise OntologyError(
                    "predicate '%s' has unknown sort '%s' (sorts=%s)" %
                    (predicate.get_name(), predicate.sort, self._sorts)
                )

        for name in self._predicates.keys():
            if name in self._sorts.keys():
                raise AmbiguousNamesException(
                    f"Expected predicate and sort names to be unique but there is both a predicate and sort named "
                    f"'{name}' in ontology '{self.name}'"
                )

    def _check_predicate_action_integrity(self):
        for action in self._actions:
            if action in self._predicates:
                raise AmbiguousNamesException(
                    "Action '%s' is also a predicate name in %s ontology." % (action, self.name)
                )

    def get_sorts(self):
        return self._sorts

    def get_predicates(self):
        return self._predicates

    def get_actions(self):
        return self._actions

    def get_ddd_specific_actions(self):
        return self._actions.difference(self.DEFAULT_ACTIONS)

    def get_action(self, name):
        if name not in self._actions:
            raise ActionDoesNotExistException("Expected one of the known sorts %s but got '%s'" % (self._actions, name))
        else:
            return Action(name, ontology_name=self.name)

    @property
    def individuals_as_objects(self):
        return [Individual(self.name, value, sort) for value, sort in self._individuals.items()]

    def get_individuals(self):
        return self._individuals

    def get_individuals_of_sort(self, expected_sort):
        return [individual for individual, sort in list(self._individuals.items()) if sort.get_name() == expected_sort]

    def add_individual(self, name, sort_as_string):
        if name in list(self._individuals.keys()):
            raise IndividualExistsException("invididual %r already exists" % name)
        self._validate_name_and_add_individual(name, sort_as_string)

    def _validate_name_and_add_individual(self, name, sort_as_string):
        self._validate_individual_name(name)
        self._individuals[name] = self.get_sort(sort_as_string)

    def ensure_individual_exists(self, name, sort_as_string):
        if name not in list(self._individuals.keys()):
            self._validate_name_and_add_individual(name, sort_as_string)

    def get_name(self):
        return self.name

    def __str__(self):
        return self.name

    def get_sort(self, name):
        if not self.has_sort(name):
            raise SortDoesNotExistException("unknown sort '%s' in %s" % (name, self))
        else:
            return self._sorts[name]

    def get_predicate(self, name):
        try:
            return self._predicates[name]
        except KeyError:
            raise OntologyError("unknown predicate %r in %s" % (name, self))

    def has_predicate(self, name):
        return name in list(self._predicates.keys())

    def has_sort(self, name):
        return name in list(self._sorts.keys())

    def is_predicate(self, object_):
        return object_ in list(self._predicates.values())

    def isPredicate(self, object_):
        warnings.warn(
            "Ontology.isPredicate() is deprecated. Use Ontology.is_predicate() instead.",
            DeprecationWarning,
            stacklevel=2
        )
        return self.is_predicate(object_)

    def individual_sort(self, value):
        def sort_of(value):
            if self._individual_is_of_enumerated_sort(value):
                return self._get_enumerated_individual_sort(value)
            return Sort.from_value(value)

        def names_of(sorts):
            return [sort.name for sort in sorts]

        sort = sort_of(value)
        if not self.predicates_contain_sort(sort.name):
            predicate_sorts = unique(names_of(self.predicate_sorts))
            raise OntologyError(
                "Expected one of the predicate sorts {} in ontology '{}', but got value '{}' of sort '{}'".format(
                    predicate_sorts, self.name, value, sort.name
                )
            )
        return sort

    def _individual_is_of_enumerated_sort(self, value):
        return value in self._individuals

    def _get_enumerated_individual_sort(self, value):
        try:
            return self._individuals[value]
        except KeyError:
            raise OntologyError("failed to get sort of unknown individual: " + str(value))

    def predicates_contain_sort(self, name):
        return name in [sort.get_name() for sort in self.predicate_sorts]

    @property
    def predicate_sorts(self):
        return unique(predicate.sort for predicate in sorted(self._predicates.values()))

    def is_action(self, value):
        return value in self._actions

    def create_lambda_abstracted_predicate_proposition(self, predicate):
        if not self.is_predicate(predicate):
            raise OntologyError("predicate %s is not valid in ontology %s" % (str(predicate), self.name))
        return LambdaAbstractedPredicateProposition(predicate, self.name)

    def create_action(self, action):
        if action not in self._actions:
            raise OntologyError("Expected one of %s but got %r" % (self._actions, action))
        return Action(action, self.name)

    def create_individual(self, value, sort=None):
        if sort is None:
            sort = self.individual_sort(value)
        value = self._normalize_and_assert_valid_individual_value(value, sort)
        return Individual(self.name, value, sort)

    def _normalize_and_assert_valid_individual_value(self, value, sort):
        if sort.is_dynamic():
            self._assume_value_of_dynamic_sort_is_valid()
            return value
        else:
            return self._normalize_and_assert_valid_individual_value_of_non_dynamic_sort(value, sort)

    def _assume_value_of_dynamic_sort_is_valid(self):
        pass

    def _normalize_and_assert_valid_individual_value_of_non_dynamic_sort(self, value, sort):
        if sort.is_builtin():
            return sort.normalize_value(value)
        else:
            self._assert_valid_individual_value_of_non_dynamic_custom_sort(value)
            return value

    def _assert_valid_individual_value_of_non_dynamic_custom_sort(self, value):
        try:
            self.individual_sort(value)
        except OntologyError as error:
            raise OntologyError("failed to create individual with unknown value: " + str(error))

    def create_negative_individual(self, value):
        try:
            sort = self.individual_sort(value)
        except OntologyError:
            raise OntologyError("failed to create negative individual with unknown value: %r" % value)
        return NegativeIndividual(self.name, value, sort)

    def create_yes_no_question(self, predicate_name, individual_value):
        predicate = self.get_predicate(predicate_name)
        individual = self.create_individual(individual_value)
        proposition = self.create_predicate_proposition(predicate, individual)
        return YesNoQuestion(proposition)

    def create_wh_question(self, predicate_name):
        predicate = self.get_predicate(predicate_name)
        lambda_abstracted_prop = self.create_lambda_abstracted_predicate_proposition(predicate)
        return WhQuestion(lambda_abstracted_prop)

    def create_predicate_proposition(self, predicate, individual, polarity=Polarity.POS):
        proposition = PredicateProposition(predicate, individual, polarity)
        return proposition



from collections import namedtuple


class QueryResultFromService(
    namedtuple("QueryResultFromService", ["value", "confidence", "grammar_entry", "is_predicted"])
):
    def __new__(cls, value, confidence, grammar_entry=None, is_predicted=False):
        return super(QueryResultFromService, cls).__new__(cls, value, confidence, grammar_entry, is_predicted)


class QueryResultFromServiceManager(namedtuple("QueryResultFromServiceManager", ["proposition", "confidence"])):
    pass



import json
import os
import tempfile

from mock import patch
import pytest

from tala.config import Config, DddConfig, BackendConfig, DeploymentsConfig, \
    UnexpectedConfigEntriesException, BackendConfigNotFoundException, DddConfigNotFoundException, \
    DeploymentsConfigNotFoundException, OptionalConfigField, \
    MandatoryConfigField, MissingMandatoryConfigEntryException


class ConfigTester(object):
    def setup_configs(self):
        self.configs = {}
        self._working_dir = os.getcwd()
        temp_dir = tempfile.mkdtemp(prefix="TestConfig")
        os.chdir(temp_dir)

    def teardown_configs(self):
        os.chdir(self._working_dir)

    def given_config_file_exists(self, *args, **kwargs):
        self._create_config_file(*args, **kwargs)

    def _create_config_file(self, ConfigClass, config, name=None):
        if name is None:
            name = ConfigClass.default_name()
        self._write_config(name, config)

    def _write_config(self, name, config):
        with open(name, 'w') as f:
            json.dump(config, f)

    def given_created_config_object(self, ConfigClass, path=None):
        self._config = ConfigClass(path)

    def given_fields(self, mock_fields, return_value):
        mock_fields.return_value = return_value

    def given_default_name(self, mock_default_name, value):
        mock_default_name.return_value = value

    def when_trying_to_read(self):
        try:
            self.when_reading()
        except UnexpectedConfigEntriesException:
            pass

    def when_reading_then_exception_is_raised_with_message_that_matches_regex(self, exception_class, expected_regex):
        with pytest.raises(exception_class, match=expected_regex):
            self.when_reading()

    def when_reading(self):
        self._result = self._config.read()

    def when_calling_default_config(self, **kwargs):
        self._result = self._config.default_config(**kwargs)

    def when_calling_default_config_then_exception_is_raised_with_message_matching_regex(
        self, exception_class, expected_regex, **kwargs
    ):
        with pytest.raises(exception_class, match=expected_regex):
            self._config.default_config(**kwargs)

    def then_result_is(self, expected):
        assert self._result == expected

    def then_config_is(self, filename, expected_config):
        with open(filename) as f:
            actual_config = json.load(f)
        assert actual_config == expected_config

    def then_wrote(self, filename, expected_config):
        with open(filename) as f:
            actual_config = json.load(f)
        assert actual_config == expected_config

    def then_result_has_key_value_pair(self, key, expected_value):
        assert self._result[key] == expected_value


class TestConfig(ConfigTester):
    def setup_method(self):
        self.setup_configs()

    def teardown_method(self):
        self.teardown_configs()

    @patch.object(Config, "fields")
    @patch.object(Config, "default_name")
    def test_read(self, mock_default_name, mock_fields):
        self.given_fields(mock_fields, {"mock_key": OptionalConfigField(default_value="mock_value")})
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_config_file_exists(Config, {"mock_key": "mock_value"})
        self.given_created_config_object(Config)
        self.when_reading()
        self.then_result_is({"mock_key": "mock_value"})

    @patch.object(Config, "fields")
    @patch.object(Config, "default_name")
    def test_read_with_unexpected_entry_raises_exception(self, mock_default_name, mock_fields):
        self.given_fields(mock_fields, {"mock_key": OptionalConfigField(default_value="mock_value")})
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_config_file_exists(Config, {"mock_key": "mock_value", "mock_unexpected_key": ""})
        self.given_created_config_object(Config)
        self.when_reading_then_exception_is_raised_with_message_that_matches_regex(
            UnexpectedConfigEntriesException, r"Parameters \['mock_unexpected_key'\] are unexpected in.*"
            r"The config was updated and the previous config was backed up in 'mock\.config\.json\.backup'\."
        )

    @patch.object(Config, "fields")
    @patch.object(Config, "default_name")
    def test_read_with_missing_optional_entries_does_not_modify_file(self, mock_default_name, mock_fields):
        self.given_fields(mock_fields, {"mock_key": OptionalConfigField(default_value="mock_value")})
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_config_file_exists(Config, {})
        self.given_created_config_object(Config)
        self.when_trying_to_read()
        self.then_config_is("mock.config.json", {})

    @patch.object(Config, "fields")
    @patch.object(Config, "default_name")
    def test_read_returns_default_value_for_missing_optional_entry(self, mock_default_name, mock_fields):
        self.given_fields(mock_fields, {"mock_key": OptionalConfigField(default_value="mock_default_value")})
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_config_file_exists(Config, {})
        self.given_created_config_object(Config)
        self.when_trying_to_read()
        self.then_result_is({"mock_key": "mock_default_value"})

    @patch.object(Config, "fields")
    @patch.object(Config, "default_name")
    def test_default_config_returns_default_value_if_not_overridden(self, mock_default_name, mock_fields):
        self.given_fields(mock_fields, {"key": OptionalConfigField(default_value="default_value")})
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_created_config_object(Config)
        self.when_calling_default_config()
        self.then_result_has_key_value_pair("key", "default_value")

    @patch.object(Config, "fields")
    @patch.object(Config, "default_name")
    def test_default_config_returns_overridden_value_if_overridden(self, mock_default_name, mock_fields):
        self.given_fields(mock_fields, {"key": OptionalConfigField(default_value="default_value")})
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_created_config_object(Config)
        self.when_calling_default_config(key="overridden_value")
        self.then_result_has_key_value_pair("key", "overridden_value")

    @patch.object(Config, "fields")
    @patch.object(Config, "default_name")
    def test_default_config_raises_exception_if_required_field_is_missing(self, mock_default_name, mock_fields):
        self.given_fields(
            mock_fields, {
                "mandatory_key": MandatoryConfigField(),
                "optional_key": OptionalConfigField(default_value="default_value")
            }
        )
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_created_config_object(Config)
        self.when_calling_default_config_then_exception_is_raised_with_message_matching_regex(
            MissingMandatoryConfigEntryException,
            r"Expected mandatory entry 'mandatory_key' but it is missing among entries \['optional_key'\]\.",
            optional_key="mock_value"
        )

    @patch.object(Config, "fields")
    @patch.object(Config, "default_name")
    def test_write_default_config(self, mock_default_name, mock_fields):
        self.given_fields(
            mock_fields, {
                "optional_key": OptionalConfigField(default_value="default_value"),
                "mandatory_key": MandatoryConfigField()
            }
        )
        self.given_default_name(mock_default_name, "mock.config.json")
        self.when_writing_default_config(Config, mandatory_key="value_for_mandatory_field")
        self.then_wrote(
            "mock.config.json", {
                "optional_key": "default_value",
                "mandatory_key": "value_for_mandatory_field"
            }
        )

    def when_writing_default_config(self, Config, **kwargs):
        Config.write_default_config(**kwargs)

    @patch.object(Config, "fields")
    @patch.object(Config, "default_name")
    def test_read_with_missing_mandatory_field_raises_exception(self, mock_default_name, mock_fields):
        self.given_fields(
            mock_fields, {
                "mandatory_key": MandatoryConfigField(),
                "optional_key": OptionalConfigField(default_value="default_value"),
            }
        )
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_config_file_exists(Config, {"optional_key": "mock_value"})
        self.given_created_config_object(Config)
        self.when_reading_then_exception_is_raised_with_message_that_matches_regex(
            MissingMandatoryConfigEntryException,
            r"Expected mandatory entry 'mandatory_key' in 'mock\.config\.json' but it is missing among "
            r"entries \['optional_key'\]\."
        )


class TestInheritance(ConfigTester):
    def setup_method(self):
        self.setup_configs()

    def teardown_method(self):
        self.teardown_configs()

    @patch.object(Config, "fields")
    def test_override_value_from_parent(self, mock_fields):
        self.given_fields(mock_fields, {"key_to_override": OptionalConfigField(default_value="mock_default_value")})
        self.given_config_file_exists(
            Config, {
                "key_to_override": "parent_value",
                "overrides": None
            }, name="parent.config.json"
        )
        self.given_config_file_exists(
            Config, {
                "overrides": "parent.config.json",
                "key_to_override": "child_value"
            }, name="child.config.json"
        )
        self.given_created_config_object(Config, "child.config.json")
        self.when_reading()
        self.then_result_is({"key_to_override": "child_value", "overrides": "parent.config.json"})

    @patch.object(Config, "fields")
    def test_keep_value_from_parent(self, mock_fields):
        self.given_fields(mock_fields, {"key_to_override": OptionalConfigField(default_value="mock_default_value")})
        self.given_config_file_exists(
            Config, {
                "key_to_override": "parent_value",
                "overrides": None
            }, name="parent.config.json"
        )
        self.given_config_file_exists(Config, {"overrides": "parent.config.json"}, name="child.config.json")
        self.given_created_config_object(Config, "child.config.json")
        self.when_reading()
        self.then_result_is({"key_to_override": "parent_value", "overrides": "parent.config.json"})

    @patch.object(Config, "fields")
    def test_read_with_missing_parent_raises_exception(self, mock_fields):
        self.given_fields(mock_fields, {})
        self.given_config_file_exists(
            BackendConfig, {"overrides": "non_existing_parent.config.json"}, name="child.config.json"
        )
        self.given_created_config_object(BackendConfig, "child.config.json")
        self.when_reading_then_exception_is_raised_with_message_that_matches_regex(
            BackendConfigNotFoundException,
            r"Expected backend config '.*non_existing_parent\.config\.json' to exist but it was not found\."
        )

    @patch.object(Config, "fields")
    @patch.object(Config, "default_name")
    def test_read_does_not_yield_exception_if_mandatory_field_is_inherited(self, mock_default_name, mock_fields):
        self.given_fields(mock_fields, {"mock_key": MandatoryConfigField()})
        self.given_default_name(mock_default_name, "mock.config.json")
        self.given_config_file_exists(
            Config, {
                "mock_key": "parent_value",
                "overrides": None
            }, name="parent.config.json"
        )
        self.given_config_file_exists(Config, {"overrides": "parent.config.json"})
        self.given_created_config_object(Config)
        self.when_reading()
        self.then_no_exception_is_raised()

    def then_no_exception_is_raised(self):
        pass


class TestDddConfig(ConfigTester):
    def setup_method(self):
        self.setup_configs()

    def teardown_method(self):
        self.teardown_configs()

    def test_default_name(self):
        assert DddConfig.default_name() == "ddd.config.json"

    def test_read_with_missing_config_file_raises_exception(self):
        self.given_created_config_object(DddConfig, "non_existing_config.json")
        self.when_reading_then_exception_is_raised_with_message_that_matches_regex(
            DddConfigNotFoundException,
            r"Expected DDD config '.*non_existing_config\.json' to exist but it was not found\."
        )


class TestDeploymentsConfig(ConfigTester):
    def setup_method(self):
        self.setup_configs()

    def teardown_method(self):
        self.teardown_configs()

    def test_default_name(self):
        assert DeploymentsConfig.default_name() == "deployments.config.json"

    def test_read_with_missing_config_file_raises_exception(self):
        self.given_created_config_object(DeploymentsConfig, "non_existing_config.json")
        self.when_reading_then_exception_is_raised_with_message_that_matches_regex(
            DeploymentsConfigNotFoundException,
            r"Expected deployments config '.*non_existing_config\.json' to exist but it was not found\."
        )

    def test_several_environments_allowed(self):
        self.given_config_file_exists(DeploymentsConfig, {"dev": "dev-url", "prod": "prod-url"})
        self.given_created_config_object(DeploymentsConfig)
        self.when_reading()
        self.then_result_is({"dev": "dev-url", "prod": "prod-url"})

    def test_get_url_when_deployment_exists_in_deployment_config(self):
        self.given_config_file_exists(DeploymentsConfig, {"dev": "dev-url", "prod": "prod-url"})
        self.given_created_config_object(DeploymentsConfig)
        self.when_getting_url_for("dev")
        self.then_url_was("dev-url")

    def when_getting_url_for(self, deployment):
        self._url = self._config.get_url(deployment)

    def then_url_was(self, expected_url):
        assert self._url == expected_url

    def test_get_url_when_deployment_is_missing_in_deployment_config(self):
        self.given_config_file_exists(DeploymentsConfig, {"dev": "dev-url", "prod": "prod-url"})
        self.given_created_config_object(DeploymentsConfig)
        self.when_getting_url_for("my.ddd.tala.cloud")
        self.then_url_was("my.ddd.tala.cloud")

    def test_get_url_when_deployment_config_is_missing(self):
        self.given_created_config_object(DeploymentsConfig)
        self.when_getting_url_for("my.ddd.tala.cloud")
        self.then_url_was("my.ddd.tala.cloud")


class TestBackendConfig(ConfigTester):
    def setup_method(self):
        self.setup_configs()

    def teardown_method(self):
        self.teardown_configs()

    def test_default_name(self):
        assert BackendConfig.default_name() == "backend.config.json"

    def test_read_with_missing_config_file_raises_exception(self):
        self.given_created_config_object(BackendConfig, "non_existing_config.json")
        self.when_reading_then_exception_is_raised_with_message_that_matches_regex(
            BackendConfigNotFoundException,
            r"Expected backend config '.*non_existing_config\.json' to exist but it was not found\."
        )

    def test_write_default_config_with_ddd(self):
        self.when_writing_default_config_with_ddd_name("my-ddd")
        self.then_wrote(
            BackendConfig.default_name(),
            BackendConfig.default_config(ddds=["my-ddd"], active_ddd="my-ddd", supported_languages="none")
        )

    def when_writing_default_config_with_ddd_name(self, ddd_name):
        BackendConfig.write_default_config(ddd_name=ddd_name)



import json

from mock import patch, Mock, call
import pytest
from requests.exceptions import RequestException

from tala.model.input_hypothesis import InputHypothesis
from tala.model.interpretation import Interpretation
from tala.model.common import Modality
from tala.model.user_move import DDDSpecificUserMove
from tala.utils import tdm_client
from tala.utils.tdm_client import TDMClient, PROTOCOL_VERSION, TDMRuntimeException


class TestTDMClient(object):
    def setup_method(self):
        self._mocked_requests = None
        self._tdm_client = None
        self._session = {"session-id": None}

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_start_session(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._when_calling_start_session()
        request_data = f'{{"version": "{PROTOCOL_VERSION}", "session": {{"session_id": "mock-id-this-should-not-appear-in-production"}}, "request": {{"start_session": {{}}}}}}'
        self._then_request_was_made_with("mock-url", data=request_data, headers={'Content-type': 'application/json'})

    def _given_mocked_requests_module(self, mock_requests):
        self._mocked_requests = mock_requests

    def _given_tdm_client_created_for(self, url):
        self._tdm_client = TDMClient(url)

    def _when_calling_start_session(self, session=None):
        self._tdm_client.start_session(session)

    def _then_request_was_made_with(self, url, headers, data):
        self._mocked_requests.post.assert_has_calls([call(url, data=data, headers=headers)])

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_start_session_with_data(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._when_calling_start_session(session={"key": "value"})
        request_data = f'{{"version": "{PROTOCOL_VERSION}", "session": {{"key": "value", "session_id": "mock-id-this-should-not-appear-in-production"}}, "request": {{"start_session": {{}}}}}}'
        self._then_request_was_made_with("mock-url", data=request_data, headers={'Content-type': 'application/json'})

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_say_with_started_session(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session-id")
        self._when_requesting_text_input("mock-utterance")
        self._then_request_was_made_with(
            "mock-url",
            data='{{"version": "{version}", "session": {{"session_id": "mock-session-id"}}, '
            '"request": {{"natural_language_input": {{'
            '"modality": "text", "utterance": "mock-utterance"}}}}}}'.format(version=PROTOCOL_VERSION),
            headers={'Content-type': 'application/json'}
        )

    def _given_session_id(self, session_id):
        self._session = {"session_id": session_id}

    def _given_requests_respond_with(self, session_id):
        response = Mock()
        json_response = '''{{
  "version": {version},
  "session": {{
    "session_id": "{session_id}"
  }},
  "output": {{
    "utterance": "mock-system-utterance",
    "expected_passivity": 5.0,
    "actions": []
  }},
  "nlu_result": {{
    "selected_utterance": "call John",
    "confidence": 0.88
  }},
  "context": {{
    "active_ddd": "my_ddd",
    "facts": {{}},
    "language": "eng"
  }}
}}'''.format(version=PROTOCOL_VERSION, session_id=session_id)
        response.json.return_value = json.loads(json_response)
        self._mocked_requests.post.return_value = response

    def _when_requesting_text_input(self, utterance, session=None):
        self._add_session_to_session_object(session)
        self._tdm_client.request_text_input(utterance, self._session)

    def _add_session_to_session_object(self, data):
        if data is not None:
            for key in data:
                self._session[key] = data[key]

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_say_with_started_session_with_data(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session-id")
        self._when_requesting_text_input("mock-utterance", session={"key": "value"})
        self._then_request_was_made_with(
            "mock-url",
            data='{{"version": "{version}", "session": {{"session_id": "mock-session-id", "key": "value"}}, '
            '"request": {{"natural_language_input": {{'
            '"modality": "text", "utterance": "mock-utterance"}}}}}}'.format(version=PROTOCOL_VERSION),
            headers={'Content-type': 'application/json'}
        )

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_requesting_speech_input_with_started_session(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session-id")
        self._when_requesting_speech_input([InputHypothesis("mock-utterance", 0.5)])
        self._then_request_was_made_with(
            "mock-url",
            data='{{"version": "{version}", "session": {{"session_id": "mock-session-id"}}, '
            '"request": {{"natural_language_input": {{'
            '"modality": "speech", '
            '"hypotheses": [{{"utterance": "mock-utterance", "confidence": 0.5}}]'
            '}}}}}}'.format(version=PROTOCOL_VERSION),
            headers={'Content-type': 'application/json'}
        )

    def _when_requesting_speech_input(self, hypotheses, session=None):
        self._add_session_to_session_object(session)
        self._tdm_client.request_speech_input(hypotheses, self._session)

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_requesting_speech_input_with_started_session_with_data(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session-id")
        self._when_requesting_speech_input([InputHypothesis("mock-utterance", 0.5)], session={"key": "value"})
        self._then_request_was_made_with(
            "mock-url",
            data='{{"version": "{version}", "session": {{"session_id": "mock-session-id", "key": "value"}}, '
            '"request": {{"natural_language_input": {{'
            '"modality": "speech", '
            '"hypotheses": [{{"utterance": "mock-utterance", "confidence": 0.5}}]'
            '}}}}}}'.format(version=PROTOCOL_VERSION),
            headers={'Content-type': 'application/json'}
        )

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_requesting_semantic_input_with_started_session(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session-id")
        self._when_requesting_semantic_input([
            Interpretation(
                [DDDSpecificUserMove(
                    "mock-ddd", "mock-expression", "mock-perception-confidence", "mock-understanding-confidence")],
                modality=Modality.SPEECH,
                utterance="mock-utterance"
            )
        ])  # yapf: disable
        self._then_request_was_made_with(
            "mock-url",
            data='{{"version": "{version}", "session": {{"session_id": "mock-session-id"}}, '
            '"request": {{"semantic_input": {{"interpretations": [{{'
            '"modality": "speech", '
            '"moves": '
            '[{{'
            '"ddd": "mock-ddd", '
            '"perception_confidence": "mock-perception-confidence", '
            '"understanding_confidence": "mock-understanding-confidence", '
            '"semantic_expression": "mock-expression"'
            '}}], '
            '"utterance": "mock-utterance"}}], "entities": []}}}}}}'.format(version=PROTOCOL_VERSION),
            headers={'Content-type': 'application/json'}
        )

    def _when_requesting_semantic_input(self, interpretations, session=None):
        self._add_session_to_session_object(session)
        self._tdm_client.request_semantic_input(interpretations, self._session)

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_requesting_semantic_input_with_started_session_with_data(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session-id")
        self._when_requesting_semantic_input([
            Interpretation([
                DDDSpecificUserMove(
                    "mock-ddd", "mock-expression", "mock-perception-confidence", "mock-understanding-confidence"
                )
            ],
                           modality=Modality.SPEECH,
                           utterance="mock-utterance")
        ],
                                             session={"key": "value"})
        self._then_request_was_made_with(
            "mock-url",
            data='{{"version": "{version}", "session": {{"session_id": "mock-session-id", "key": "value"}}, '
            '"request": {{"semantic_input": {{"interpretations": [{{'
            '"modality": "speech", '
            '"moves": '
            '[{{'
            '"ddd": "mock-ddd", '
            '"perception_confidence": "mock-perception-confidence", '
            '"understanding_confidence": "mock-understanding-confidence", '
            '"semantic_expression": "mock-expression"'
            '}}], '
            '"utterance": "mock-utterance"}}], "entities": []}}}}}}'.format(version=PROTOCOL_VERSION),
            headers={'Content-type': 'application/json'}
        )

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_passivity_with_started_session(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session-id")
        self._when_calling_passivity()
        self._then_request_was_made_with(
            "mock-url",
            data='{{"version": "{version}", "session": {{"session_id": "mock-session-id"}}, '
            '"request": {{"passivity": {{}}}}}}'.format(version=PROTOCOL_VERSION),
            headers={'Content-type': 'application/json'}
        )

    def _when_calling_passivity(self, session=None):
        self._add_session_to_session_object(session)
        self._tdm_client.request_passivity(self._session)

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_passivity_with_started_session_and_data(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session-id")
        self._when_calling_passivity(session={"key": "value"})
        self._then_request_was_made_with(
            "mock-url",
            data='{{"version": "{version}", "session": {{"session_id": "mock-session-id", "key": "value"}}, '
            '"request": {{"passivity": {{}}}}}}'.format(version=PROTOCOL_VERSION),
            headers={'Content-type': 'application/json'}
        )

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_start_session_with_error_response(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_requests_error_with("mock-error")
        self._when_starting_session_then_exception_was_raised(TDMRuntimeException, "mock-error")

    def _when_starting_session_then_exception_was_raised(self, expected_exception, expected_message):
        with pytest.raises(expected_exception, match=expected_message):
            self._tdm_client.start_session()

    def _given_requests_error_with(self, description):
        response = Mock()
        json_response = '''{{
  "version": {version},
  "session": {{
    "session_id": "mock-session"
  }},
  "error": {{
    "description": "{description}"
  }}
}}'''.format(version=PROTOCOL_VERSION, description=description)
        response.json.return_value = json.loads(json_response)
        self._mocked_requests.post.return_value = response

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_unexpected_status_code(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_response_status_check_raises(RequestException, "mock-error")
        self._when_starting_session_then_exception_was_raised(RequestException, "mock-error")

    def _given_response_status_check_raises(self, ExceptionClass, message):
        self._mocked_requests.post.return_value.raise_for_status.side_effect = ExceptionClass(message)

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_say_with_error_response(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session")
        self._given_requests_error_with("mock-error")
        self._when_requesting_text_input_then_exception_was_raised("mock-utterance", TDMRuntimeException, "mock-error")

    def _when_requesting_text_input_then_exception_was_raised(self, utterance, expected_exception, expected_message):
        with pytest.raises(expected_exception, match=expected_message):
            self._tdm_client.request_text_input(utterance, self._session)

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_request_passivity_with_error_response(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session")
        self._given_requests_error_with("mock-error")
        self._when_requesting_passivity_then_exception_was_raised(TDMRuntimeException, "mock-error")

    def _when_requesting_passivity_then_exception_was_raised(self, expected_exception, expected_message):
        with pytest.raises(expected_exception, match=expected_message):
            self._tdm_client.request_passivity(self._session)

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_request_speech_input_with_error_response(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session")
        self._given_requests_error_with("mock-error")
        self._when_requesting_speech_input_then_exception_was_raised([InputHypothesis("mock-utterance", 0.5)],
                                                                     TDMRuntimeException, "mock-error")

    def _when_requesting_speech_input_then_exception_was_raised(self, hypotheses, expected_exception, expected_message):
        with pytest.raises(expected_exception, match=expected_message):
            self._tdm_client.request_speech_input(hypotheses, self._session)

    @patch("{}.requests".format(tdm_client.__name__), autospec=True)
    def test_request_semantic_input_with_error_response(self, mock_requests):
        self._given_mocked_requests_module(mock_requests)
        self._given_tdm_client_created_for("mock-url")
        self._given_session_id("mock-session")
        self._given_requests_error_with("mock-error")
        self._when_requesting_semantic_input_then_exception_was_raised([
            Interpretation(
                [DDDSpecificUserMove(
                    "mock-ddd", "mock-expression", "mock-perception-confidence", "mock-understanding-confidence")],
                modality=Modality.SPEECH,
                utterance="mock-utterance"
            )
        ], TDMRuntimeException, "mock-error")  # yapf: disable

    def _when_requesting_semantic_input_then_exception_was_raised(
        self, interpretations, expected_exception, expected_message
    ):
        with pytest.raises(expected_exception, match=expected_message):
            self._tdm_client.request_semantic_input(interpretations, self._session)



import codecs
from fnmatch import fnmatch
import warnings

from tala.testing.interactions.compiler import InteractionTestCompiler
from tala.testing.interactions.named_test import InteractionTest


class InteractionTestingFile:
    def __init__(self, filename, tests):
        self.filename = filename
        self.tests = tests

    def add(self, name, interaction_events):
        test = InteractionTest(self.filename, name, interaction_events)
        self.tests.append(test)

    def __iter__(self):
        for test in self.tests:
            yield test

    def __getitem__(self, index):
        return self.tests[index]

    def __len__(self):
        return len(self.tests)

    def __repr__(self):
        return "InteractionTestingFile(filename=%s, tests=%r)" % (self.filename, self.tests)

    def pop_first(self):
        return self.tests.pop(0)

    def select_tests(self, patterns):
        warnings.warn(
            "This method modifies the tests internally in a destructive way. Use filter_tests_by_name() "
            "instead, which creates and returns a new list with the tests that match.", DeprecationWarning
        )
        self.tests = self.filter_tests_by_name(patterns)

    def filter_tests_by_name(self, patterns):
        filtered_tests = []
        for test in self.tests:
            for selected_test in patterns:
                if fnmatch(test.name, selected_test):
                    filtered_tests.append(test)
        return filtered_tests

    @staticmethod
    def _load_tests_from_file(path):
        with codecs.open(path, 'r', encoding='utf-8') as f:
            return InteractionTestCompiler().compile_interaction_tests(path, f)

    @staticmethod
    def from_path(path):
        tests = InteractionTestingFile._load_tests_from_file(path)
        return InteractionTestingFile(path, tests)



import warnings

from tala.model.move import RequestMove, AskMove
from tala.model import speaker
from tala.model.semantic_object import SemanticObject, SemanticObjectWithContent
from tala.utils.as_semantic_expression import AsSemanticExpressionMixin
from tala.utils.unicodify import unicodify

PERFORM = "PERFORM_GOAL"
RESOLVE = "RESOLVE_GOAL"


class Goal(SemanticObject, AsSemanticExpressionMixin):
    def __init__(self, goal_type, target):
        self._goal_type = goal_type
        self._target = target
        self._background = None

    def is_goal(self):
        return True

    def is_perform_goal(self):
        return False

    def is_resolve_goal(self):
        return False

    def is_handle_goal(self):
        return False

    def is_top_goal(self):
        return False

    @property
    def type_(self):
        return self._goal_type

    @property
    def type(self):
        warnings.warn("Goal.type is deprecated. Use Goal.type_ instead.", DeprecationWarning, stacklevel=2)
        return self.type_

    @property
    def target(self):
        return self._target

    def __ne__(self, other):
        return not (self == other)

    def __eq__(self, other):
        try:
            return (other.target == self.target and other.type_ == self.type_)
        except AttributeError:
            return False

    def __hash__(self):
        return hash((self._goal_type, self._target))

    def __repr__(self):
        return "%s(%s, %s)" % (Goal.__name__, self._goal_type, self._target)

    def __str__(self):
        return repr(self)

    def set_background(self, background):
        self._background = background

    @staticmethod
    def goal_filter(goal_type):
        return lambda goal: goal.type_ == goal_type

    @staticmethod
    def goal_target_filter(target_speaker):
        return lambda goal: goal.target == target_speaker


class GoalWithSemanticContent(Goal, SemanticObjectWithContent):
    def __init__(self, goal_type, target, content):
        Goal.__init__(self, goal_type, target)
        SemanticObjectWithContent.__init__(self, content)
        self._content = content

    @property
    def content(self):
        return self._content

    def get_content(self):
        warnings.warn(
            "GoalWithSemanticContent.get_content() is deprecated. Use GoalWithSemanticContent.content instead.",
            DeprecationWarning,
            stacklevel=2
        )
        return self._content

    def is_goal_with_semantic_content(self):
        return True

    def __hash__(self):
        try:
            return hash((self._goal_type, self._target, self._content, self.ontology_name))
        except NotImplementedError:
            return hash((self._goal_type, self._target, self._content))

    def __repr__(self):
        try:
            return "%s(%s, %s, %s, %s)" % (
                GoalWithSemanticContent.__name__, self.ontology_name, self._goal_type, self._target, self._content
            )
        except NotImplementedError:
            return "%s(%s, %s, %s)" % (GoalWithSemanticContent.__name__, self._goal_type, self._target, self._content)

    def __eq__(self, other):
        def ontologies_match():
            if self.is_ontology_specific():
                return other.is_ontology_specific() and other.ontology_name == self.ontology_name
            else:
                return not other.is_ontology_specific()

        try:
            return other.is_goal() \
                and other.has_semantic_content() \
                and other.content == self.content \
                and other.target == self.target \
                and other.type_ == self.type_ \
                and ontologies_match()
        except AttributeError:
            pass
        except NotImplementedError:
            pass
        return False

    def as_move(self):
        raise NotImplementedError()


class Perform(GoalWithSemanticContent):
    def __init__(self, action, target=speaker.SYS):
        assert action.is_action()
        GoalWithSemanticContent.__init__(self, PERFORM, target, action)

    def is_perform_goal(self):
        return True

    @staticmethod
    def filter():
        return Goal.goal_filter(PERFORM)

    def get_action(self):
        warnings.warn("Goal.get_action() is deprecated. Use Goal.action instead.", DeprecationWarning, stacklevel=2)
        return self.action

    @property
    def action(self):
        return self.content

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return "perform(%s)" % self.action

    def as_move(self):
        return RequestMove(self.content)

    def is_top_goal(self):
        return self.action.is_top_action()


class PerformGoal(Perform):
    pass


class Resolve(GoalWithSemanticContent):
    def __init__(self, question, target):
        assert question.is_question()
        GoalWithSemanticContent.__init__(self, RESOLVE, target, question)

    def is_resolve_goal(self):
        return True

    @property
    def question(self):
        return self.content

    @property
    def issue(self):
        return self.content

    def get_question(self):
        warnings.warn("Goal.get_question() is deprecated. Use Goal.question instead.", DeprecationWarning, stacklevel=2)
        return self.content

    @staticmethod
    def filter():
        return Goal.goal_filter(RESOLVE)

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        result = ""
        if self._target == speaker.USR:
            result += "resolve_user"
        else:
            result += "resolve"
        result += "(%s" % self.question
        if self._background:
            result += ", %s" % unicodify(self._background)
        result += ")"
        return result

    def as_move(self):
        return AskMove(self.content)


class ResolveGoal(Resolve):
    pass



import unittest

from mock import Mock

from tala.ddd.ddd_component_manager import DDDComponentManager, \
    DDDSpecificComponentsAlreadyExistsException
from tala.ddd.services.service_interface import ServiceInterface
from tala.model.ddd import DDD
from tala.model.domain import Domain
from tala.model.ontology import Ontology
from tala.ddd.ddd_specific_components import DDDSpecificComponents
from tala.ddd.services.parameters.retriever import ParameterRetriever


class TestManagerForDDDSpecificComponents(unittest.TestCase):
    def setUp(self):
        self.domain = None
        self.ontology = None
        self._mock_device_handler = None
        self._ddd_specific_components = None
        self._second_ddd_specific_components = None
        self._ddd_component_manager = None

    def _create_ddd(self, name="mockup_ddd"):
        return DDD(name, ontology=self.ontology, domain=self.domain, service_interface=Mock(spec=ServiceInterface))

    def _create_ddd_specific_components(self, ddd, device_class=None):
        return DDDSpecificComponents(ddd, parameter_retriever=Mock(spec=ParameterRetriever), parser=None)

    def test_get_domain(self):
        self._given_empty_manager_for_ddd_specific_components()
        self._given_mocked_ontology("MockupOntology")
        self._given_mocked_domain("MockupDomain")
        self._given_ddd_specific_components()
        self._given_ddd_specific_components_added()
        self._when_getting_domain("MockupDomain")
        self._then_it_equals_mocked_domain()

    def _given_mocked_domain(self, name="MockupDomain"):
        self.domain = self._create_mock_domain(name)

    def _create_mock_domain(self, name):
        domain = Mock(spec=Domain)
        domain.get_name.return_value = name
        domain.goals = {}
        return domain

    def _given_ddd_specific_components(self, name=None, device_class=None):
        ddd = self._create_ddd(name)
        self._ddd_specific_components = self._create_ddd_specific_components(ddd, device_class)

    def _given_second_ddd_specific_components(self, name=None, device_class=None):
        ddd = self._create_ddd(name)
        self._second_ddd_specific_components = self._create_ddd_specific_components(ddd, device_class)

    def _when_getting_domain(self, name):
        self._result = self._ddd_component_manager.get_domain(name)

    def _then_it_equals_mocked_domain(self):
        self.assertEqual(self._result, self.domain)

    def test_get_ontology(self):
        self._given_empty_manager_for_ddd_specific_components()
        self._given_mocked_ontology("MockupOntology")
        self._given_mocked_domain("MockupDomain")
        self._given_ddd_specific_components()
        self._given_ddd_specific_components_added()
        self._when_getting_ontology("MockupOntology")
        self._then_it_equals_mocked_ontology()

    def _given_mocked_ontology(self, name="MockupOntology"):
        self.ontology = self._create_mock_ontology(name)

    def _create_mock_ontology(self, name):
        ontology = Mock(spec=Ontology)
        ontology.get_individuals.return_value = {}
        ontology.get_name.return_value = name
        return ontology

    def _when_getting_ontology(self, name):
        self._result = self._ddd_component_manager.get_ontology(name)

    def _then_it_equals_mocked_ontology(self):
        self.assertEqual(self._result, self.ontology)

    def test_adding_duplicate_yields_exception(self):
        self._given_empty_manager_for_ddd_specific_components()
        self._given_mocked_ontology("MockupOntology")
        self._given_mocked_domain("MockupDomain")
        self._given_ddd_specific_components("ddd")
        self._given_ddd_specific_components_added()
        self._given_second_ddd_specific_components("ddd")
        self._when_adding_second_ddd_specific_components_again_then_exception_is_raised()

    def _when_adding_second_ddd_specific_components_again_then_exception_is_raised(self):
        with self.assertRaises(DDDSpecificComponentsAlreadyExistsException):
            self._given_second_ddd_specific_components_added()

    def _given_empty_manager_for_ddd_specific_components(self):
        self._ddd_component_manager = DDDComponentManager()

    def _given_ddd_specific_components_added(self):
        self._ddd_component_manager.add(self._ddd_specific_components)

    def _given_second_ddd_specific_components_added(self):
        self._ddd_component_manager.add(self._second_ddd_specific_components)

    def _then_sorted_results_is(self, expected_result):
        self.assertEqual(expected_result, sorted(self._result))

    def test_get_ddd(self):
        self._given_mocked_ontology()
        self._given_mocked_domain()
        self._given_empty_manager_for_ddd_specific_components()
        self._given_ddd_specific_components("ddd")
        self._given_ddd_specific_components_added()
        self._when_getting_ddd_specific_components_for("ddd")
        self._then_ddd_specific_components_are_returned()

    def _when_getting_ddd_specific_components_for(self, ddd):
        self._result = self._ddd_component_manager.get_ddd_specific_components(ddd)

    def _then_ddd_specific_components_are_returned(self):
        self.assertEqual(self._result, self._ddd_specific_components)

    def test_reset_ddd_resets_ontology(self):
        self._given_mocked_ontology()
        self._given_mocked_domain()
        self._given_empty_manager_for_ddd_specific_components()
        self._given_ddd_specific_components("ddd")
        self._given_ddd_specific_components_added()
        self._given_individual_added_to_ontology()
        self._when_ddd_is_reset_for("ddd")
        self._then_get_ddd_returns_ddd_with_unmodified_ontology("ddd")

    def _given_individual_added_to_ontology(self):
        self.ontology.add_individual("athens", "city")

    def _when_ddd_is_reset_for(self, ddd):
        self._ddd_component_manager.reset_components_of_ddd(ddd)

    def _then_get_ddd_returns_ddd_with_unmodified_ontology(self, ddd_name):
        ddd = self._ddd_component_manager.get_ddd_specific_components(ddd_name)
        self.assertEqual(0, len(ddd.ontology.get_individuals()))

    def _then_result_is(self, expected):
        self.assertEqual(expected, self._result)



import copy
import unittest

from mock import Mock

from tala.model.error import OntologyError
from tala.model.goal import PerformGoal
from tala.model import speaker
from tala.model.ontology import Ontology
from tala.model.polarity import Polarity
from tala.model.question import YesNoQuestion
from tala.model.proposition import PreconfirmationProposition, ResolvednessProposition, PrereportProposition, ServiceActionTerminatedProposition, RejectedPropositions, PropositionSet, ServiceResultProposition, Proposition, QuitProposition, MuteProposition, UnderstandingProposition, UnmuteProposition, PredicateProposition, GoalProposition, KnowledgePreconditionProposition
from tala.model.sort import StringSort, ImageSort
from tala.testing.lib_test_case import LibTestCase
from tala.model.image import Image
from tala.model.set import Set


class TestPropositionAsCondition(unittest.TestCase):
    def setUp(self):
        pass

    def test_proposition_truth_value_given_empty_facts(self):
        self.given_a_proposition(Proposition("some_type"))
        self.given_a_set_of_facts([])
        self.when_proposition_evaluated_with_facts()
        self.then_result_is(False)

    def given_a_proposition(self, proposition):
        self.proposition = proposition

    def given_a_set_of_facts(self, facts):
        self.facts = Set()
        for fact in facts:
            self.facts.add(fact)

    def when_proposition_evaluated_with_facts(self):
        self.result = self.proposition.is_true_given_proposition_set(self.facts)

    def then_result_is(self, truth_value):
        self.assertEqual(truth_value, self.result)

    def test_proposition_truth_value_false_given_non_empty_facts(self):
        self.given_a_proposition(Proposition("some_type"))
        self.given_a_set_of_facts([Proposition("other_type")])
        self.when_proposition_evaluated_with_facts()
        self.then_result_is(False)

    def test_proposition_truth_value_true_given_non_empty_facts_with_same_proposition(self):
        self.given_a_proposition(Proposition("some_type"))
        self.given_a_set_of_facts([Proposition("other_type"), Proposition("some_type")])
        self.when_proposition_evaluated_with_facts()
        self.then_result_is(True)


class PropositionTests(LibTestCase, unittest.TestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def test_quit_proposition_string(self):
        quit_prop = QuitProposition()
        self.assertEqual("QUIT", str(quit_prop))

    def test_quit_proposition_get_type(self):
        quit_prop = QuitProposition()
        self.assertEqual("QUIT", quit_prop.get_type())

    def test_mute_proposition_string(self):
        mute_prop = MuteProposition()
        self.assertEqual("MUTE", str(mute_prop))

    def test_unmute_proposition_string(self):
        unmute_prop = UnmuteProposition()
        self.assertEqual("UNMUTE", str(unmute_prop))

    def test_quit_proposition_equality(self):
        quit_proposition_1 = QuitProposition()
        quit_proposition_2 = QuitProposition()
        self.assertEqual(quit_proposition_1, quit_proposition_2)

    def test_understanding_proposition_getters(self):
        proposition = self.proposition_dest_city_paris
        und = UnderstandingProposition(speaker.USR, proposition)
        self.assertEqual(speaker.USR, und.get_speaker())
        self.assertEqual(proposition, und.get_content())

    def test_understanding_proposition_to_string(self):
        proposition = self.proposition_dest_city_paris
        und = UnderstandingProposition(speaker.USR, proposition)
        self.assertEqual("und(USR, dest_city(paris))", str(und))


class ServiceResultPropositionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.action = "mockup_action"
        self.arguments = ["mockup_arguments"]
        self.proposition = ServiceResultProposition(
            self.ontology_name, self.action, self.arguments, "mock_perform_result"
        )

    def test_is_service_result_proposition(self):
        self.assertTrue(self.proposition.is_service_result_proposition())

    def test_getters(self):
        self.assertEqual(self.action, self.proposition.get_service_action())
        self.assertEqual(self.arguments, self.proposition.get_arguments())
        self.assertEqual("mock_perform_result", self.proposition.get_result())

    def test_equality(self):
        identical_proposition = ServiceResultProposition(
            self.ontology_name, self.action, self.arguments, "mock_perform_result"
        )
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(self.proposition, identical_proposition)

    def test_inequality_due_to_result(self):
        non_identical_proposition = ServiceResultProposition(
            self.ontology_name, self.action, self.arguments, "other_perform_result"
        )
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.proposition, non_identical_proposition)

    def test_inequality_due_to_action(self):
        other_action = "other_service_action"
        non_identical_proposition = ServiceResultProposition(
            self.ontology_name, other_action, self.arguments, True, None
        )
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.proposition, non_identical_proposition)

    def test_unicode(self):
        self.assertEqual(
            "ServiceResultProposition(mockup_action, ['mockup_arguments'], mock_perform_result)", str(self.proposition)
        )

    def test_hashable(self):
        {self.proposition}


class PropositionSetTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.positive_prop_set = PropositionSet([self.proposition_dest_city_london, self.proposition_dest_city_paris])
        self.negative_prop_set = PropositionSet([self.proposition_dest_city_london, self.proposition_dest_city_paris],
                                                polarity=Polarity.NEG)

    def test_equality(self):
        identical_prop_set = PropositionSet([self.proposition_dest_city_london, self.proposition_dest_city_paris])
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(self.positive_prop_set, identical_prop_set)

    def test_inequality_due_to_propositions(self):
        non_identical_prop_set = PropositionSet([self.proposition_dest_city_london])
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.positive_prop_set, non_identical_prop_set)

    def test_inequality_due_to_polarity(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.negative_prop_set, self.positive_prop_set)

    def test_inequality_due_to_element_order(self):
        non_identical_prop_set = PropositionSet([self.proposition_dest_city_paris, self.proposition_dest_city_london])
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.positive_prop_set, non_identical_prop_set)

    def test_is_positive(self):
        self.assertTrue(self.positive_prop_set.is_positive())
        self.assertFalse(self.negative_prop_set.is_positive())

    def test_get_polarity(self):
        self.assertEqual(Polarity.POS, self.positive_prop_set.get_polarity())
        self.assertEqual(Polarity.NEG, self.negative_prop_set.get_polarity())

    def test_positive_propositionset_unicode(self):
        self.assertEqual("set([dest_city(london), dest_city(paris)])", str(self.positive_prop_set))

    def test_negative_propositionset_unicode(self):
        self.assertEqual("~set([dest_city(london), dest_city(paris)])", str(self.negative_prop_set))

    def test_is_understanding_proposition_false(self):
        self.assertFalse(self.positive_prop_set.is_understanding_proposition())

    def test_is_proposition_set(self):
        self.assertTrue(self.positive_prop_set.is_proposition_set())
        self.assertFalse(self.buy_action.is_proposition_set())

    def test_get_predicate(self):
        self.assertEqual(self.predicate_dest_city, self.positive_prop_set.predicate)

    def test_repr(self):
        self.assertEqual(
            "PropositionSet([PredicateProposition("
            "Predicate('dest_city', CustomSort('city', True), None, False), "
            "Individual('london', CustomSort('city', True)), "
            "'POS', False), "
            "PredicateProposition("
            "Predicate('dest_city', CustomSort('city', True), None, False), "
            "Individual('paris', CustomSort('city', True)"
            "), 'POS', False)], 'POS')", repr(self.positive_prop_set)
        )

    def test_is_ontology_specific_with_some_ontology_specific_propositions(self):
        self.given_proposition_set_with([
            self._create_mocked_ontology_specific_proposition("an ontology"),
            self._create_mocked_proposition(),
        ])
        self.when_retrieving_is_ontology_specific()
        self.then_result_is(False)

    def _create_mocked_ontology_specific_proposition(self, ontology_name):
        mock = Mock(spec=PreconfirmationProposition)
        mock.is_ontology_specific.return_value = True
        mock.ontology_name = ontology_name
        return mock

    def _create_mocked_proposition(self):
        mock = Mock(spec=Proposition)
        mock.is_ontology_specific.return_value = False
        return mock

    def given_proposition_set_with(self, propositions):
        self._proposition_set = PropositionSet(propositions)

    def when_retrieving_is_ontology_specific(self):
        self._actual_result = self._proposition_set.is_ontology_specific()

    def test_is_ontology_specific_with_only_ontology_specific_propositions_of_different_ontologies(self):
        self.given_proposition_set_with([
            self._create_mocked_ontology_specific_proposition("an ontology"),
            self._create_mocked_ontology_specific_proposition("another ontology"),
        ])
        self.when_retrieving_is_ontology_specific()
        self.then_result_is(False)

    def test_is_ontology_specific_with_only_ontology_specific_propositions_of_same_ontologies(self):
        self.given_proposition_set_with([
            self._create_mocked_ontology_specific_proposition("an ontology"),
            self._create_mocked_ontology_specific_proposition("an ontology"),
        ])
        self.when_retrieving_is_ontology_specific()
        self.then_result_is(True)

    def test_is_ontology_specific_with_no_ontology_specific_propositions(self):
        self.given_proposition_set_with([
            self._create_mocked_proposition(),
            self._create_mocked_proposition(),
        ])
        self.when_retrieving_is_ontology_specific()
        self.then_result_is(False)

    def test_is_ontology_specific_without_propositions(self):
        self.given_proposition_set_with([])
        self.when_retrieving_is_ontology_specific()
        self.then_result_is(False)


class ServiceActionTerminatedTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.service_action = "MakeReservation"
        self.done = ServiceActionTerminatedProposition(self.ontology_name, self.service_action)

    def test_done_proposition_is_done_prop(self):
        self.assertTrue(self.done.is_service_action_terminated_proposition())

    def test_done_proposition_contains_action(self):
        self.assertEqual(self.service_action, self.done.get_service_action())

    def test_done_proposition_equality(self):
        identical_done_prop = ServiceActionTerminatedProposition(self.ontology_name, self.service_action)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(identical_done_prop, self.done)

    def test_done_proposition_inequality(self):
        inequal_done_prop = ServiceActionTerminatedProposition(self.ontology_name, "hskjhs")
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(inequal_done_prop, self.done)

    def test_done_proposition_unicode(self):
        expected_str = "service_action_terminated(MakeReservation)"
        self.assertEqual(expected_str, str(self.done))

    def test_hashing(self):
        {self.done}

    def test_repr(self):
        self.assertEqual(
            "ServiceActionTerminatedProposition('mockup_ontology', 'MakeReservation', 'POS')", repr(self.done)
        )


class RejectionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.rejected_combination = PropositionSet([self.proposition_dest_city_paris])
        self.rejected = RejectedPropositions(self.rejected_combination)

    def test_rejected_proposition_is_rejected_prop(self):
        self.assertTrue(self.rejected.is_rejected_proposition())

    def test_creation_with_reason(self):
        reason = "ssdlfkjslkf"
        rejected = RejectedPropositions(self.rejected_combination, reason=reason)
        self.assertEqual(self.rejected_combination, rejected.get_rejected_combination())
        self.assertEqual(reason, rejected.get_reason())

    def test_rejected_with_reason_unicode(self):
        reason = "ssdlfkjslkf"
        rejected = RejectedPropositions(self.rejected_combination, reason=reason)
        string = "rejected(%s, %s)" % (self.rejected_combination, reason)
        self.assertEqual(string, str(rejected))

    def test_get_rejected_combination(self):
        self.assertEqual(self.rejected_combination, self.rejected.get_rejected_combination())

    def test_rejected_proposition_unicode(self):
        rejected_as_string = "rejected(%s)" % self.rejected_combination
        self.assertEqual(rejected_as_string, str(self.rejected))

    def test_equality(self):
        identical_rejected_prop = RejectedPropositions(PropositionSet([self.proposition_dest_city_paris]))
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(identical_rejected_prop, self.rejected)

    def test_inequality_due_to_rejected_combination(self):
        other_combination = "hskjhs"
        inequal_rejected_prop = RejectedPropositions(PropositionSet(other_combination))
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(inequal_rejected_prop, self.rejected)

    def test_inequality_due_to_reason(self):
        inequal_rejected_prop = RejectedPropositions(self.rejected_combination, reason="some_reason")
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(inequal_rejected_prop, self.rejected)

    def test_hashable(self):
        prop_set = {self.rejected}
        self.assertTrue(self.rejected in prop_set)


class PrereportPropositionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.service_action = "MakeReservation"
        self.empty_arguments = []
        self.prereport = self._create_with_empty_param_list()

    def _create_with_empty_param_list(self):
        confirmation_proposition = PrereportProposition(self.ontology_name, self.service_action, self.empty_arguments)
        return confirmation_proposition

    def test_type(self):
        self.assertTrue(self.prereport.is_prereport_proposition())

    def test_get_service_action(self):
        self.assertEqual(self.service_action, self.prereport.get_service_action())

    def test_get_arguments(self):
        self.assertEqual(self.empty_arguments, self.prereport.get_arguments())

    def test_proposition_unicode(self):
        self.assertEqual("prereported(MakeReservation, [])", str(self.prereport))

    def test_equality(self):
        identical_prereport = PrereportProposition(self.ontology_name, self.service_action, self.empty_arguments)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(self.prereport, identical_prereport)

    def test_inequality_due_to_action(self):
        other_action = "CacelTransaction"
        non_identical_proposition = PrereportProposition(self.ontology_name, other_action, self.empty_arguments)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.prereport, non_identical_proposition)


class PreconfirmationTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.service_action = "MakeReservation"
        self.arguments = [self.proposition_dest_city_paris, self.proposition_dept_city_london]
        self.preconfirmation = PreconfirmationProposition(self.ontology_name, self.service_action, self.arguments)

    def test_type(self):
        self.assertTrue(self.preconfirmation.is_preconfirmation_proposition())

    def test_get_service_action(self):
        self.assertEqual(self.service_action, self.preconfirmation.get_service_action())

    def test_get_arguments(self):
        self.assertEqual(self.arguments, self.preconfirmation.get_arguments())

    def test_positive_proposition_unicode(self):
        self.assertEqual(
            "preconfirmed(MakeReservation, [dest_city(paris), dept_city(london)])", str(self.preconfirmation)
        )

    def test_negative_proposition_unicode(self):
        negative_preconfirmation = self.preconfirmation.negate()
        self.assertEqual(
            "~preconfirmed(MakeReservation, [dest_city(paris), dept_city(london)])", str(negative_preconfirmation)
        )

    def test_equality(self):
        identical_preconfirmation = PreconfirmationProposition(self.ontology_name, self.service_action, self.arguments)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(self.preconfirmation, identical_preconfirmation)

    def test_inequality_due_to_other_class(self):
        object_of_other_class = ServiceActionTerminatedProposition(self.ontology_name, self.service_action)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.preconfirmation, object_of_other_class)

    def test_inequality_due_to_action(self):
        other_action = "CancelTransaction"
        non_identical_proposition = PreconfirmationProposition(self.ontology_name, other_action, self.arguments)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.preconfirmation, non_identical_proposition)

    def test_inequality_due_to_polarity(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(
            self.preconfirmation, self.preconfirmation.negate()
        )

    def test_inequality_due_to_arguments_in_other_order(self):
        other_arguments = [self.proposition_dept_city_london, self.proposition_dest_city_paris]
        other_preconfirmation = PreconfirmationProposition(self.ontology_name, self.service_action, other_arguments)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.preconfirmation, other_preconfirmation)

    def test_default_polarity_is_positive(self):
        self.assertTrue(self.preconfirmation.is_positive())

    def test_is_positive_false_for_negative_proposition(self):
        self.assertFalse(self.preconfirmation.negate().is_positive())

    def test_hashable(self):
        {self.preconfirmation}


class ResolvednessPropositionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.issue = self.price_question
        self.proposition = ResolvednessProposition(self.issue)

    def test_get_issue(self):
        self.assertEqual(self.issue, self.proposition.get_issue())

    def test_unicode(self):
        self.assertEqual("resolved(?X.price(X))", str(self.proposition))

    def test_equality(self):
        identical_proposition = ResolvednessProposition(self.issue)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(self.proposition, identical_proposition)

    def test_inequality_due_to_issue(self):
        other_issue = self.dest_city_question
        non_identical_proposition = ResolvednessProposition(other_issue)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.proposition, non_identical_proposition)

    def test_hashable(self):
        {self.proposition}


class QuestionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def test_is_preconfirmation_question_true_for_yes_no_questions_about_confirmation(self):
        preconfirmation_question = YesNoQuestion(
            PreconfirmationProposition(self.ontology_name, "mockup_service_action", [])
        )
        self.assertTrue(preconfirmation_question.is_preconfirmation_question())

    def test_is_preconfirmation_question_false_for_yes_no_questions_about_non_preconfirmations(self):
        non_preconfirmation_question = YesNoQuestion(self.proposition_dest_city_paris)
        self.assertFalse(non_preconfirmation_question.is_preconfirmation_question())

    def test_is_preconfirmation_question_false_for_non_yes_no_questions(self):
        self.assertFalse(self.dest_city_question.is_preconfirmation_question())


class PredicatePropositionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def test_getters(self):
        self.assertEqual(self.predicate_dest_city, self.proposition_dest_city_paris.predicate)
        self.assertEqual(self.individual_paris, self.proposition_dest_city_paris.individual)

    def test_positive_unicode(self):
        self.assertEqual("dest_city(paris)", str(self.proposition_dest_city_paris))

    def test_negative_unicode(self):
        self.assertEqual("~dest_city(paris)", str(self.proposition_not_dest_city_paris))

    def test_equality(self):
        proposition = self.proposition_dest_city_paris
        identical_proposition = copy.copy(self.proposition_dest_city_paris)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(proposition, identical_proposition)

    def test_inequality_due_to_individual(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(
            self.proposition_dest_city_paris, self.proposition_dest_city_london
        )

    def test_inequality_due_to_polarity(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(
            self.proposition_dest_city_paris, self.proposition_not_dest_city_paris
        )

    def test_default_polarity_is_positive(self):
        proposition = PredicateProposition(self.predicate_dest_city, self.individual_paris)
        self.assertTrue(proposition.is_positive())

    def test_default_polarity_is_positive_with_help_method(self):
        proposition = PredicateProposition(self.predicate_dest_city, self.individual_paris)
        self.assertTrue(proposition.is_positive())

    def test_negative_proposition(self):
        proposition = PredicateProposition(self.predicate_dest_city, self.individual_paris, Polarity.NEG)
        self.assertFalse(proposition.is_positive())

    def test_create_illegal_proposition(self):
        with self.assertRaises(OntologyError):
            predicate = self.ontology.get_predicate("destination")
            individual = self.ontology.create_individual("car")
            PredicateProposition(predicate, individual)

    def test_predicate(self):
        self.assertEqual("dest_city", self.proposition_dest_city_paris.predicate.get_name())

    def test_value(self):
        self.assertEqual("paris", self.proposition_dest_city_paris.individual.value)

    def test_hashable(self):
        {self.proposition_dest_city_paris}

    def test_nullary(self):
        proposition = PredicateProposition(self.predicate_need_visa)
        self.assertEqual(None, proposition.individual)
        self.assertEqual("need_visa", str(proposition))

    def test_is_predicate_proposition_true(self):
        self.assertTrue(self.proposition_dest_city_paris.is_predicate_proposition())

    def test_is_predicate_proposition_false(self):
        self.assertFalse(self.individual_paris.is_predicate_proposition())

    def test_predicate_proposition_with_sortal_mismatch_yields_exception(self):
        individual_of_wrong_type = self.ontology.create_individual("paris")
        with self.assertRaises(OntologyError):
            PredicateProposition(self.predicate_price, individual_of_wrong_type)


class GoalPropositionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.proposition = GoalProposition(PerformGoal(self.buy_action))
        self.negative_proposition = GoalProposition(PerformGoal(self.buy_action), polarity=Polarity.NEG)

    def test_equality(self):
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(
            GoalProposition(PerformGoal(self.buy_action)), GoalProposition(PerformGoal(self.buy_action))
        )

    def test_inequality_due_to_goal(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(
            GoalProposition(PerformGoal(self.buy_action)), GoalProposition(PerformGoal(self.top_action))
        )

    def test_inequality_due_to_polarity(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(
            GoalProposition(PerformGoal(self.buy_action), polarity=Polarity.POS),
            GoalProposition(PerformGoal(self.buy_action), polarity=Polarity.NEG)
        )

    def test_hashable(self):
        {self.proposition}

    def test_positive_proposition_unicode(self):
        self.assertEqual("goal(perform(buy))", str(self.proposition))

    def test_negative_proposition_unicode(self):
        self.assertEqual("~goal(perform(buy))", str(self.negative_proposition))


class StringPropositionTests(LibTestCase):
    def setUp(self):
        self._create_ontology()

    def _create_ontology(self):
        self.ontology_name = "mockup_ontology"
        predicates = {self._create_predicate("number_to_call", StringSort())}
        sorts = set()
        individuals = {}
        actions = set()
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, actions)

    def test_str_with_string_argument(self):
        predicate_number_to_call = self.ontology.get_predicate("number_to_call")
        individual = self.ontology.create_individual('"123"')
        string_proposition = PredicateProposition(predicate_number_to_call, individual)
        self.assertEqual('number_to_call("123")', str(string_proposition))


class ImagePropositionTests(LibTestCase):
    def setUp(self):
        self._create_ontology()

    def _create_ontology(self):
        self.ontology_name = "mockup_ontology"
        predicates = {self._create_predicate("map_to_show", ImageSort())}
        sorts = set()
        individuals = {}
        actions = set()
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, actions)

    def test_str_with_string_argument(self):
        map_to_show = self.ontology.get_predicate("map_to_show")
        individual = self.ontology.create_individual(Image("http://mymap.com/map.png"))
        image_proposition = PredicateProposition(map_to_show, individual)
        self.assertEqual('map_to_show(image("http://mymap.com/map.png"))', str(image_proposition))


class KnowledgePreconditionPropositionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def test_equality(self):
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(
            KnowledgePreconditionProposition("mock_question", Polarity.POS),
            KnowledgePreconditionProposition("mock_question", Polarity.POS)
        )

    def test_inequality_due_to_question(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(
            KnowledgePreconditionProposition("mock_question_1", Polarity.POS),
            KnowledgePreconditionProposition("mock_question_2", Polarity.POS)
        )

    def test_inequality_due_to_polarity(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(
            KnowledgePreconditionProposition("mock_question", Polarity.POS),
            KnowledgePreconditionProposition("mock_question", Polarity.NEG)
        )



import json
from typing import Sequence, Mapping  # noqa: F401

import requests

from tala.model.input_hypothesis import InputHypothesis  # noqa: F401
from tala.model.interpretation import Interpretation  # noqa: F401
from tala.model.event_notification import EventNotification  # noqa: F401
from tala.utils.await_endpoint import await_endpoint
from tala.utils.observable import Observable

PROTOCOL_VERSION = "3.4"


class InvalidResponseError(Exception):
    pass


class TDMRuntimeException(Exception):
    pass


class TDMClient(Observable):
    def __init__(self, url: str) -> None:
        super(TDMClient, self).__init__()
        self._url = url

    def wait_to_start(self, timeout=3):
        url = self._url.replace("/interact", "/health")
        await_endpoint(url, timeout)

    def request_text_input(self, utterance: str, session: Mapping = None) -> Mapping:
        def _create_text_input_request(session, utterance):
            return {
                "version": PROTOCOL_VERSION,
                "session": session,
                "request": {
                    "natural_language_input": {
                        "modality": "text",
                        "utterance": utterance,
                    }
                }
            }

        request = _create_text_input_request(session, utterance)
        response = self._make_request(request)
        return response

    def request_speech_input(self, hypotheses: Sequence[InputHypothesis], session: Mapping = None) -> Mapping:
        def _create_speech_input_request(session, hypotheses):
            return {
                "version": PROTOCOL_VERSION,
                "session": session,
                "request": {
                    "natural_language_input": {
                        "modality": "speech",
                        "hypotheses": [{
                            "utterance": hypothesis.utterance,
                            "confidence": hypothesis.confidence,
                        } for hypothesis in hypotheses],
                    }
                }
            }  # yapf: disable

        request = _create_speech_input_request(session, hypotheses)
        response = self._make_request(request)
        return response

    def request_semantic_input(
        self, interpretations: Sequence[Interpretation], session: Mapping = None, entities: Sequence = None
    ) -> Mapping:
        def _create_semantic_input_request(session, interpretations, entities):
            if entities is None:
                entities = []
            return {
                "version": PROTOCOL_VERSION,
                "session": session,
                "request": {
                    "semantic_input": {
                        "interpretations": [interpretation.as_dict() for interpretation in interpretations],
                        "entities": entities,
                    }
                }
            }  # yapf: disable
        request = _create_semantic_input_request(session, interpretations, entities)
        response = self._make_request(request)
        return response

    def request_passivity(self, session: Mapping = None) -> Mapping:
        request = {"version": PROTOCOL_VERSION, "session": session, "request": {"passivity": {}}}
        response = self._make_request(request)
        return response

    def request_event_notification(self, notification: EventNotification, session: Mapping = None) -> Mapping:
        def _create_event_notification_request(session, notification):
            return {
                "version": PROTOCOL_VERSION,
                "session": session,
                "request": {
                    "event": {
                        "name": notification.action,
                        "status": notification.status,
                        "parameters": notification.parameters
                    }
                }
            }  # yapf: disable

        request = _create_event_notification_request(session, notification)
        response = self._make_request(request)
        return response

    def start_session(self, session_data: Mapping = None) -> Mapping:
        session_object = session_data or {}
        if "session_id" not in session_object:
            session_object["session_id"] = "mock-id-this-should-not-appear-in-production"
        request = {"version": PROTOCOL_VERSION, "session": session_object, "request": {"start_session": {}}}
        response = self._make_request(request)
        return response

    def make_request(self, *args, **kwargs):
        return self._make_request(*args, **kwargs)

    def _make_request(self, request_body):
        data_as_json = json.dumps(request_body)
        headers = {'Content-type': 'application/json'}
        response_object = requests.post(self._url, data=data_as_json, headers=headers)
        response_object.raise_for_status()
        try:
            response = response_object.json()
        except ValueError:
            raise InvalidResponseError(f"Expected a valid JSON response but got '{response_object}'.")

        if "error" in response:
            description = response["error"]["description"]
            raise TDMRuntimeException(description)
        return response



from mock import patch
import pytest

from tala.model.ontology import Ontology, SortDoesNotExistException
from tala.model.predicate import Predicate
from tala.model.sort import Sort, StringSort, ImageSort, WebviewSort, BooleanSort, RealSort, CustomSort, IntegerSort, \
    DateTimeSort, InvalidValueException, BuiltinSortRepository, UndefinedSort, DomainSort, PersonNameSort, BOOLEAN, \
    INTEGER, DATETIME, PERSON_NAME, REAL, STRING, IMAGE, DOMAIN, WEBVIEW, UnsupportedValue
import tala.model.sort
from tala.model.image import Image
from tala.model.webview import Webview
from tala.model.date_time import DateTime
from tala.model.person_name import PersonName
from tala.testing.utils import EqualityAssertionTestCaseMixin


class SortTestCase(object):
    def setup_method(self):
        self._sort = None
        self.ontology_name = None

    def _create_predicate(self, *args, **kwargs):
        return Predicate(self.ontology_name, *args, **kwargs)

    def when_normalize_value(self, value):
        self._actual_result = self._sort.normalize_value(value)

    def when_normalize_value_then_exception_is_raised(self, value, expected_exception, expected_message):
        with pytest.raises(expected_exception, match=expected_message):
            self._sort.normalize_value(value)

    def then_result_is(self, expected_result):
        assert expected_result == self._actual_result

    def given_sort(self, sort):
        self._sort = sort

    def when_call_value_as_basic_type(self, value):
        self._actual_result = self._sort.value_as_basic_type(value)

    def when_call_value_from_basic_type(self, basic_value):
        self._actual_result = self._sort.value_from_basic_type(basic_value)

    def when_call_value_as_json_object(self, value):
        self._actual_result = self._sort.value_as_json_object(value)

    def when_call_grammar_entry(self, value, language_code):
        self._actual_result = self._sort.grammar_entry(value, language_code)


class TestSort(SortTestCase, EqualityAssertionTestCaseMixin):
    def setup_method(self):
        self._create_ontology()

    def _create_ontology(self):
        self.ontology_name = "mockup_ontology"
        predicates = set()
        sorts = {
            CustomSort(self.ontology_name, "city", dynamic=False),
            CustomSort(self.ontology_name, "city_type", dynamic=False),
        }
        individuals = {}
        actions = set()
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, actions)

    def test_createSort(self):
        sort = self.ontology.get_sort("city")
        assert "city" == sort.get_name()

    def test_createSort_yields_exception_for_unknown_sort(self):
        with pytest.raises(SortDoesNotExistException):
            self.ontology.get_sort("sdkjfhskdjf")

    def test_sort_unicode(self):
        assert "Sort('city', dynamic=False)" == str(CustomSort(self.ontology_name, "city"))

    def test_sort_equality(self):
        sort1 = self.ontology.get_sort("city")
        sort2 = self.ontology.get_sort("city")
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(sort1, sort2)

    def test_sort_inequality(self):
        sort1 = self.ontology.get_sort("city")
        sort2 = self.ontology.get_sort("city_type")
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(sort1, sort2)

    def test_sorts_hashable(self):
        {self.ontology.get_sort("city"), RealSort()}

    def test_value_as_basic_type_returns_value_by_default(self):
        self.given_sort(CustomSort(self.ontology_name, "city", dynamic=False))
        self.when_call_value_as_basic_type("mock_value")
        self.then_result_is("mock_value")

    def test_value_from_basic_type_returns_basic_value_by_default(self):
        self.given_sort(CustomSort(self.ontology_name, "city", dynamic=False))
        self.when_call_value_from_basic_type("mock_value")
        self.then_result_is("mock_value")

    @pytest.mark.parametrize(
        "value,expected_sort", [(123.50, RealSort()), (123, IntegerSort()),
                                (Image("http://www.internet.org/image.png"), ImageSort()),
                                (Webview("http://maps.com/map.html"), WebviewSort())]
    )
    def test_from_value_returns_expected_sort(self, value, expected_sort):
        self.when_call_from_value(value)
        self.then_result_is(expected_sort)

    def when_call_from_value(self, value):
        self._actual_result = Sort.from_value(value)

    def test_from_value_raises_exception_for_unsupported_value(self):
        with pytest.raises(UnsupportedValue, match="Expected a supported value but got 'None'"):
            self.when_call_from_value(None)

    @patch("{}.Sort.value_as_basic_type".format(tala.model.sort.__name__))
    def test_value_as_json_object_uses_value_as_basic_type(self, mock_value_as_basic_type):
        self.given_a_sort()
        self.when_call_value_as_json_object("mock_value")
        self.then_result_is({"value": mock_value_as_basic_type.return_value})

    def given_a_sort(self, name="mock_name"):
        self._sort = Sort(name)


class TestStringSort(SortTestCase):
    def setup_method(self):
        self._create_ontology()
        self._sort = StringSort()

    def _create_ontology(self):
        self.ontology_name = "mockup_ontology"
        predicates = {self._create_predicate("number_to_call", StringSort())}
        sorts = set()
        individuals = {}
        actions = set()
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, actions)

    def test_string_individual(self):
        individual = self.ontology.create_individual('"a string"')
        assert "a string" == individual.value
        assert StringSort() == individual.sort

    def test_string_individual_unicode(self):
        individual = self.ontology.create_individual('"a unicode string"')
        assert "a unicode string" == individual.value
        assert StringSort() == individual.sort

    def test_string_individual_unicode_method(self):
        individual = self.ontology.create_individual('"a string"')
        assert '"a string"' == str(individual)

    def test_normalize_base_case(self):
        self.when_normalize_value("a string")
        self.then_result_is("a string")

    def test_exception_when_normalize_non_string_value(self):
        self.when_normalize_value_then_exception_is_raised(
            123, InvalidValueException, "Expected a string value but got 123 of type int."
        )


class TestImageSort(SortTestCase):
    def setup_method(self):
        self._create_ontology()
        self._sort = ImageSort()

    def _create_ontology(self):
        self.ontology_name = "mockup_ontology"
        predicates = {self._create_predicate("map", ImageSort())}
        sorts = set()
        individuals = {}
        actions = set()
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, actions)

    def test_create_individual(self):
        individual = self.ontology.create_individual(Image("http://image.com/image.png"))
        assert Image("http://image.com/image.png") == individual.value
        assert individual.sort.is_image_sort()

    def test_is_not_dynamic(self):
        image_sort = ImageSort()
        assert not image_sort.is_dynamic()

    def test_normalize_static_url_of_image_mime_type(self):
        self.when_normalize_value(Image("http://image.com/image.png"))
        self.then_result_is(Image("http://image.com/image.png"))

    @patch("%s.mimetypes" % tala.model.sort.__name__)
    @patch("%s.urllib" % tala.model.sort.__name__)
    def test_normalize_dynamic_url_of_image_mime_type(self, mock_urllib, mock_mimetypes):
        self.given_get_mime_type_of_url_returns(mock_mimetypes, mock_urllib, "image/jpeg")
        self.given_urllib_parsed_path_is(mock_urllib, "/generate_image")
        self.when_normalize_value(Image("http://mock.domain/generate_image"))
        self.then_result_is(Image("http://mock.domain/generate_image"))

    def given_get_mime_type_of_url_returns(self, mock_mimetypes, mock_urllib, mime_type):
        mock_urllib.request.urlretrieve.return_value = "mock_filepath", "mock_headers"
        mock_mimetypes.guess_type.return_value = mime_type, None

    def given_urllib_parsed_path_is(self, mock_urllib, path):
        mock_urllib.parse.urlparse.return_value.path = path

    def test_exception_when_normalize_non_url(self):
        self.when_normalize_value_then_exception_is_raised(
            Image("non_url"), InvalidValueException, "Expected an image URL but got 'non_url'."
        )

    def test_exception_when_normalize_non_image(self):
        self.when_normalize_value_then_exception_is_raised(
            "non_image", InvalidValueException, "Expected an image object but got 'non_image'."
        )

    def test_exception_when_normalize_static_url_of_non_image_mime_type(self):
        self.when_normalize_value_then_exception_is_raised(
            Image("http://nonimage.com/nonimage.html"), InvalidValueException,
            "Expected an image URL but got 'http://nonimage.com/nonimage.html'."
        )

    @patch(f"{tala.model.sort.__name__}.mimetypes")
    @patch(f"{tala.model.sort.__name__}.urllib")
    def test_exception_when_normalize_dynamic_url_of_non_image_mime_type(self, mock_urllib, mock_mimetypes):
        self.given_get_mime_type_of_url_returns(mock_mimetypes, mock_urllib, "text/html")
        self.given_urllib_parsed_path_is(mock_urllib, "/generate_html")
        self.when_normalize_value_then_exception_is_raised(
            Image("http://mock.domain/generate_html"), InvalidValueException,
            "Expected an image URL but got 'http://mock.domain/generate_html'."
        )

    def test_value_as_basic_type(self):
        self.when_call_value_as_basic_type(Image("http://image.com/image.png"))
        self.then_result_is("http://image.com/image.png")

    def test_value_from_basic_type(self):
        self.when_call_value_from_basic_type("http://image.com/image.png")
        self.then_result_is(Image("http://image.com/image.png"))


class TestWebviewSort(SortTestCase):
    def setup_method(self):
        self._create_ontology()
        self._sort = WebviewSort()

    def _create_ontology(self):
        self.ontology_name = "mockup_ontology"
        predicates = {self._create_predicate("interactive_map", WebviewSort())}
        sorts = set()
        individuals = {}
        actions = set()
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, actions)

    def test_create_individual(self):
        individual = self.ontology.create_individual(Webview("http://maps.com/map.html"))
        assert Webview("http://maps.com/map.html") == individual.value
        assert WebviewSort() == individual.sort

    def test_is_not_dynamic(self):
        image_sort = WebviewSort()
        assert not image_sort.is_dynamic()

    def test_normalize_base_case(self):
        self.when_normalize_value(Webview("http://maps.com/map.html"))
        self.then_result_is(Webview("http://maps.com/map.html"))

    def test_exception_when_normalize_non_webview_value(self):
        self.when_normalize_value_then_exception_is_raised(
            "non_webview_value", InvalidValueException, "Expected a webview object but got 'non_webview_value'."
        )

    def test_exception_when_normalize_non_url(self):
        self.when_normalize_value_then_exception_is_raised(
            Webview("non_url"), InvalidValueException, "Expected a webview URL but got 'non_url'."
        )

    def test_value_as_basic_type(self):
        self.when_call_value_as_basic_type(Webview("http://maps.com/map.html"))
        self.then_result_is("http://maps.com/map.html")

    def test_value_from_basic_type(self):
        self.when_call_value_from_basic_type("http://maps.com/map.html")
        self.then_result_is(Webview("http://maps.com/map.html"))


class TestBooleanSort(SortTestCase):
    def setup_method(self):
        self._create_ontology()
        self._sort = BooleanSort()

    def _create_ontology(self):
        self.ontology_name = "mockup_ontology"
        predicates = {self._create_predicate("need_visa", BooleanSort())}
        sorts = set()
        individuals = {}
        actions = set()
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, actions)

    def test_created_individual_has_expected_value(self):
        self.given_created_individual_with_argument(True)
        self.when_call_getValue_of_individual()
        self.then_result_is(True)

    def given_created_individual_with_argument(self, value):
        self._individual = self.ontology.create_individual(value)

    def when_call_getValue_of_individual(self):
        self._actual_result = self._individual.value

    def test_created_individual_has_expected_sort(self):
        self.given_created_individual_with_argument(True)
        self.when_call_getSort_of_individual()
        self.then_result_is(BooleanSort())

    def when_call_getSort_of_individual(self):
        self._actual_result = self._individual.sort

    def test_normalize_true(self):
        self.when_normalize_value(True)
        self.then_result_is(True)

    def test_normalize_false(self):
        self.when_normalize_value(False)
        self.then_result_is(False)

    def test_exception_when_normalize_non_boolean_value(self):
        self.when_normalize_value_then_exception_is_raised(
            "non_boolean_value", InvalidValueException, "Expected a boolean value but got 'non_boolean_value'."
        )


class TestIntegerSort(SortTestCase):
    def setup_method(self):
        self._sort = IntegerSort()

    def test_normalize_base_case(self):
        self.when_normalize_value("123")
        self.then_result_is(123)

    def test_exception_when_normalize_non_integer_value(self):
        self.when_normalize_value_then_exception_is_raised(
            "non_integer_value", InvalidValueException, "Expected an integer value but got 'non_integer_value'."
        )


class TestRealSort(SortTestCase):
    def setup_method(self):
        self._sort = RealSort()

    def test_normalize_base_case(self):
        self.when_normalize_value("123.4")
        self.then_result_is(123.4)

    def test_exception_when_normalize_non_real_value(self):
        self.when_normalize_value_then_exception_is_raised(
            "non_real_value", InvalidValueException, "Expected a real-number value but got 'non_real_value'."
        )


class TestDateTimeSort(SortTestCase):
    def setup_method(self):
        self._sort = DateTimeSort()

    def test_normalize_base_case(self):
        self.when_normalize_value(DateTime("2018-03-20T22:00:00.000Z"))
        self.then_result_is(DateTime("2018-03-20T22:00:00.000Z"))

    def test_exception_when_normalize_datetime_object_with_non_iso8601_value(self):
        self.when_normalize_value_then_exception_is_raised(
            DateTime("non_iso8601_value"), InvalidValueException,
            "Expected a datetime value in ISO 8601 format but got 'non_iso8601_value'."
        )

    def test_exception_when_normalize_non_datetime_value(self):
        self.when_normalize_value_then_exception_is_raised(
            "non_datetime_value", InvalidValueException, "Expected a datetime object but got 'non_datetime_value'."
        )

    def test_value_as_basic_type(self):
        self.when_call_value_as_basic_type(DateTime("2018-03-20T22:00:00.000Z"))
        self.then_result_is("2018-03-20T22:00:00.000Z")

    def test_value_from_basic_type(self):
        self.when_call_value_from_basic_type("2018-03-20T22:00:00.000Z")
        self.then_result_is(DateTime("2018-03-20T22:00:00.000Z"))


class TestPersonName(SortTestCase):
    def setup_method(self):
        self._sort = PersonNameSort()

    def test_normalize_base_case(self):
        self.when_normalize_value(PersonName("John"))
        self.then_result_is(PersonName("John"))

    def test_normalize_from_string(self):
        self.when_normalize_value("John")
        self.then_result_is(PersonName("John"))

    def test_exception_when_normalize_non_person_name_value(self):
        self.when_normalize_value_then_exception_is_raised(
            10, InvalidValueException, "Expected a person name object but got 10 of type int."
        )

    def test_value_as_basic_type(self):
        self.when_call_value_as_basic_type(PersonName("John"))
        self.then_result_is("John")

    def test_value_from_basic_type(self):
        self.when_call_value_from_basic_type("John")
        self.then_result_is(PersonName("John"))


class TestBuiltinSortRepository(object):
    @pytest.mark.parametrize("name", [BOOLEAN, INTEGER, DATETIME, PERSON_NAME, REAL, STRING, IMAGE, DOMAIN, WEBVIEW])
    def test_has_sort_true(self, name):
        self.when_invoking_has_sort(name)
        self.then_result_is(True)

    def when_invoking_has_sort(self, name):
        self._actual_result = BuiltinSortRepository.has_sort(name)

    def then_result_is(self, expected_result):
        assert expected_result == self._actual_result

    def test_has_sort_false(self):
        self.when_invoking_has_sort("undefined_sort")
        self.then_result_is(False)

    @pytest.mark.parametrize(
        "name,expected_class", [(BOOLEAN, BooleanSort), (INTEGER, IntegerSort), (DATETIME, DateTimeSort),
                                (REAL, RealSort), (STRING, StringSort), (IMAGE, ImageSort), (DOMAIN, DomainSort),
                                (WEBVIEW, WebviewSort), (PERSON_NAME, PersonNameSort)]
    )
    def test_get_sort_successful(self, name, expected_class):
        self.when_invoking_get_sort(name)
        self.then_result_is_instance_of(expected_class)

    def when_invoking_get_sort(self, name):
        self._actual_result = BuiltinSortRepository.get_sort(name)

    def then_result_is_instance_of(self, expected_class):
        assert expected_class == self._actual_result.__class__

    def test_get_sort_unsuccessful(self):
        self.when_invoking_get_sort_then_exception_is_raised(
            "undefined_sort", UndefinedSort, "Expected a built-in sort but got 'undefined_sort'."
        )

    def when_invoking_get_sort_then_exception_is_raised(self, name, expected_class, expected_message):
        with pytest.raises(expected_class) as excinfo:
            BuiltinSortRepository.get_sort(name)
        assert expected_message == str(excinfo.value)



import pytest

from tala.ddd.files import ddd_files


class TestFiles:
    def setup_method(self):
        self._path = None
        self._result = None

    @pytest.mark.skip
    def test_ddd_files(self):
        self.given_path("tdm/ddds/person_name")
        self.when_fetching_ddd_files()
        self.then_returns([
            'tdm/ddds/person_name/__init__.py', 'tdm/ddds/person_name/ddd.config.json',
            'tdm/ddds/person_name/device.py', 'tdm/ddds/person_name/domain.xml',
            'tdm/ddds/person_name/grammar/__init__.py', 'tdm/ddds/person_name/grammar/grammar_eng.xml',
            'tdm/ddds/person_name/http_service.py', 'tdm/ddds/person_name/ontology.xml',
            'tdm/ddds/person_name/service_interface.xml', 'tdm/ddds/person_name/test/interaction_tests_eng.txt'
        ])

    def given_path(self, path):
        self._path = path

    def when_fetching_ddd_files(self):
        self._result = ddd_files(self._path)

    def then_returns(self, expected_result):
        actual_result = [str(actual) for actual in self._result]
        assert expected_result == sorted(actual_result), f"Expected {expected_result} but got {self._result}"
